import torch
import torchvision

from ..models.model import Discriminator, Generator
from .backends import *

if TENSORBOARD_LOGGING == 1:
    from tensorboardX import SummaryWriter
if VISDOM_LOGGING == 1:
    import visdom

__all__ = [
    "Visualize",
    "LossVisualize",
    "MetricVisualize",
    "GradientVisualize",
    "ImageVisualize",
]


class Visualize(object):
    r"""Base class for all Visualizations.

    Args:
        visualize_list (list, optional): List of the functions needed for visualization.
        visdom_port (int, optional): Port to log using ``visdom``. The visdom server needs to be
            manually started at this port else an error will be thrown and the code will crash.
            This is ignored if ``VISDOM_LOGGING`` is ``0``.
        log_dir (str, optional): Directory where TensorboardX should store the logs. This is
            ignored if ``TENSORBOARD_LOGGING`` is ``0``.
        writer (tensorboardX.SummaryWriter, optonal): Send a `SummaryWriter` if you
            don't want to start a new SummaryWriter.
    """

    def __init__(self, visualize_list, visdom_port=8097, log_dir=None, writer=None):
        self.logs = {}
        for item in visualize_list:
            name = type(item).__name__
            self.logs[name] = []
        self.step = 1
        if TENSORBOARD_LOGGING == 1:
            self._build_tensorboard(log_dir, writer)
        if VISDOM_LOGGING == 1:
            self._build_visdom(visdom_port)

    def _build_tensorboard(self, log_dir, writer):
        r"""Starts the tensorboard logging utilities.

        Args:
            log_dir (str, optional): Directory where TensorboardX should store the logs.
            writer (tensorboardX.SummaryWriter, optonal): Send a `SummaryWriter` if you
                don't want to start a new SummaryWriter.
        """
        self.writer = SummaryWriter(log_dir) if writer is None else writer

    def _build_visdom(self, port):
        r"""Starts the visdom logging utilities.

        Args:
            port (int, optional): Port to log using ``visdom``. A deafult server is started at port
                ``8097``. So manually a new server has to be started if the post is changed.
        """
        self.vis = visdom.Visdom(port=port)

    def step_update(self):
        r"""Helper function which updates the step at the end of
        one print iteration.
        """
        self.step += 1

    def log_tensorboard(self):
        r"""Tensorboard logging function. Needs to be defined in the subclass

        :raises NotImplementedError:
        """
        raise NotImplementedError

    def log_console(self):
        r"""Console logging function. Needs to be defined in the subclass

        :raises NotImplementedError:
        """
        raise NotImplementedError

    def log_visdom(self):
        r"""Visdom logging function. Needs to be defined in the subclass

        :raises NotImplementedError:
        """
        raise NotImplementedError

    def __call__(
        self,
        *args,
        lock_console=False,
        lock_tensorboard=False,
        lock_visdom=False,
        **kwargs
    ):
        if not lock_console and CONSOLE_LOGGING == 1:
            self.log_console(*args, **kwargs)
        if not lock_tensorboard and TENSORBOARD_LOGGING == 1:
            self.log_tensorboard(*args, **kwargs)
        if not lock_visdom and VISDOM_LOGGING == 1:
            self.log_visdom(*args, **kwargs)
        self.step_update()


class LossVisualize(Visualize):
    r"""This class provides the Visualizations for Generator and Discriminator Losses.

    Args:
        visualize_list (list, optional): List of the functions needed for visualization.
        visdom_port (int, optional): Port to log using ``visdom``. The visdom server needs to be
            manually started at this port else an error will be thrown and the code will crash.
            This is ignored if ``VISDOM_LOGGING`` is ``0``.
        log_dir (str, optional): Directory where TensorboardX should store the logs. This is
            ignored if ``TENSORBOARD_LOGGING`` is ``0``.
        writer (tensorboardX.SummaryWriter, optonal): Send a `SummaryWriter` if you
            don't want to start a new SummaryWriter.
    """

    def log_tensorboard(self, running_losses):
        r"""Tensorboard logging function. This function logs the following:

        - ``Running Discriminator Loss``
        - ``Running Generator Loss``
        - ``Running Losses``
        - Loss Values of the individual Losses.

        Args:
            running_losses (dict): A dict with 2 items namely, ``Running Discriminator Loss``,
                and ``Running Generator Loss``.
        """
        self.writer.add_scalar(
            "Running Discriminator Loss",
            running_losses["Running Discriminator Loss"],
            self.step,
        )
        self.writer.add_scalar(
            "Running Generator Loss",
            running_losses["Running Generator Loss"],
            self.step,
        )
        self.writer.add_scalars("Running Losses", running_losses, self.step)
        for name, value in self.logs.items():
            val = value[-1]
            if type(val) is tuple:
                self.writer.add_scalar(
                    "Losses/{}-Generator".format(name), val[0], self.step
                )
                self.writer.add_scalar(
                    "Losses/{}-Discriminator".format(name), val[1], self.step
                )
            else:
                self.writer.add_scalar("Losses/{}".format(name), val, self.step)

    def log_console(self, running_losses):
        r"""Console logging function. This function logs the mean ``generator`` and ``discriminator``
        losses.

        Args:
            running_losses (dict): A dict with 2 items namely, ``Running Discriminator Loss``,
                and ``Running Generator Loss``.
        """
        for name, val in running_losses.items():
            print("Mean {} : {}".format(name, val))

    def log_visdom(self, running_losses):
        r"""Visdom logging function. This function logs the following:

        - ``Running Discriminator Loss``
        - ``Running Generator Loss``
        - ``Running Losses``
        - Loss Values of the individual Losses.

        Args:
            running_losses (dict): A dict with 2 items namely, ``Running Discriminator Loss``,
                and ``Running Generator Loss``.
        """
        self.vis.line(
            [running_losses["Running Discriminator Loss"]],
            [self.step],
            win="Running Discriminator Loss",
            update="append",
            opts=dict(
                title="Running Discriminator Loss",
                xlabel="Time Step",
                ylabel="Running Loss",
            ),
        )
        self.vis.line(
            [running_losses["Running Generator Loss"]],
            [self.step],
            win="Running Generator Loss",
            update="append",
            opts=dict(
                title="Running Generator Loss",
                xlabel="Time Step",
                ylabel="Running Loss",
            ),
        )
        self.vis.line(
            [
                [
                    running_losses["Running Discriminator Loss"],
                    running_losses["Running Generator Loss"],
                ]
            ],
            [self.step],
            win="Running Losses",
            update="append",
            opts=dict(
                title="Running Losses",
                xlabel="Time Step",
                ylabel="Running Loss",
                legend=["Discriminator", "Generator"],
            ),
        )
        for name, value in self.logs.items():
            val = value[-1]
            if type(val) is tuple:
                name1 = "{}-Generator".format(name)
                name2 = "{}-Discriminator".format(name)
                self.vis.line(
                    [val[0]],
                    [self.step],
                    win=name1,
                    update="append",
                    opts=dict(title=name1, xlabel="Time Step", ylabel="Loss Value"),
                )
                self.vis.line(
                    [val[1]],
                    [self.step],
                    win=name2,
                    update="append",
                    opts=dict(title=name2, xlabel="Time Step", ylabel="Loss Value"),
                )
            else:
                self.vis.line(
                    [val],
                    [self.step],
                    win=name,
                    update="append",
                    opts=dict(title=name, xlabel="Time Step", ylabel="Loss Value"),
                )

    def __call__(self, trainer, **kwargs):
        running_generator_loss = (
            trainer.loss_information["generator_losses"]
            / trainer.loss_information["generator_iters"]
        )
        running_discriminator_loss = (
            trainer.loss_information["discriminator_losses"]
            / trainer.loss_information["discriminator_iters"]
        )
        running_losses = {
            "Running Discriminator Loss": running_discriminator_loss,
            "Running Generator Loss": running_generator_loss,
        }
        super(LossVisualize, self).__call__(running_losses, **kwargs)


class MetricVisualize(Visualize):
    r"""This class provides the Visualizations for Metrics.

    Args:
        visualize_list (list, optional): List of the functions needed for visualization.
        visdom_port (int, optional): Port to log using ``visdom``. The visdom server needs to be
            manually started at this port else an error will be thrown and the code will crash.
            This is ignored if ``VISDOM_LOGGING`` is ``0``.
        log_dir (str, optional): Directory where TensorboardX should store the logs. This is
            ignored if ``TENSORBOARD_LOGGING`` is ``0``.
        writer (tensorboardX.SummaryWriter, optonal): Send a `SummaryWriter` if you
            don't want to start a new SummaryWriter.
    """

    def log_tensorboard(self):
        r"""Tensorboard logging function. This function logs the values of the individual metrics.
        """
        for name, value in self.logs.items():
            self.writer.add_scalar("Metrics/{}".format(name), value[-1], self.step)

    def log_console(self):
        r"""Console logging function. This function logs the mean metrics.
        """
        for name, val in self.logs.items():
            print("{} : {}".format(name, val[-1]))

    def log_visdom(self):
        r"""Visdom logging function. This function logs the values of the individual metrics.
        """
        for name, value in self.logs.items():
            self.vis.line(
                [value[-1]],
                [self.step],
                win=name,
                update="append",
                opts=dict(title=name, xlabel="Time Step", ylabel="Metric Value"),
            )


class GradientVisualize(Visualize):
    r"""This class provides the Visualizations for the Gradients.

    Args:
        visualize_list (list, optional): List of the functions needed for visualization.
        visdom_port (int, optional): Port to log using ``visdom``. The visdom server needs to be
            manually started at this port else an error will be thrown and the code will crash.
            This is ignored if ``VISDOM_LOGGING`` is ``0``.
        log_dir (str, optional): Directory where TensorboardX should store the logs. This is
            ignored if ``TENSORBOARD_LOGGING`` is ``0``.
        writer (tensorboardX.SummaryWriter, optonal): Send a `SummaryWriter` if you
            don't want to start a new SummaryWriter.
    """

    def __init__(self, visualize_list, visdom_port=8097, log_dir=None, writer=None):
        if visualize_list is None or len(visualize_list) == 0:
            raise Exception("Gradient Visualizer requires list of model names")
        self.logs = {}
        for item in visualize_list:
            self.logs[item] = [0.0]
        self.step = 1
        if TENSORBOARD_LOGGING == 1:
            self._build_tensorboard(log_dir, writer)
        if VISDOM_LOGGING == 1:
            self._build_visdom(visdom_port)

    def log_tensorboard(self, name):
        r"""Tensorboard logging function. This function logs the values of the individual gradients.

        Args:
            name (str): Name of the model whose gradients are to be logged.
        """
        self.writer.add_scalar(
            "Gradients/{}".format(name),
            self.logs[name][len(self.logs[name]) - 1],
            self.step,
        )

    def log_console(self, name):
        r"""Console logging function. This function logs the mean gradients.

        Args:
            name (str): Name of the model whose gradients are to be logged.
        """
        print(
            "{} Gradients : {}".format(name, self.logs[name][len(self.logs[name]) - 1])
        )

    def log_visdom(self, name):
        r"""Visdom logging function. This function logs the values of the individual gradients.

        Args:
            name (str): Name of the model whose gradients are to be logged.
        """
        self.vis.line(
            [self.logs[name][len(self.logs[name]) - 1]],
            [self.step],
            win=name,
            update="append",
            opts=dict(title=name, xlabel="Time Step", ylabel="Gradient"),
        )

    def update_grads(self, name, model, eps=1e-5):
        r"""Updates the gradient logs.

        Args:
            name (str): Name of the model.
            model (torch.nn.Module): Either a ``torchgan.models.Generator`` or a
                ``torchgan.models.Discriminator`` or their subclass.
            eps (float, optional): Tolerance value.
        """
        gradsum = 0.0
        for p in model.parameters():
            if p.grad is not None:
                gradsum += torch.sum(p.grad ** 2).clone().item()
        if gradsum > eps:
            self.logs[name][len(self.logs[name]) - 1] += gradsum
            model.zero_grad()

    def report_end_epoch(self):
        r"""Prints to the console at the end of the epoch.
        """
        if CONSOLE_LOGGING == 1:
            for key, val in self.logs.items():
                print("{} Mean Gradients : {}".format(key, sum(val) / len(val)))

    def __call__(self, trainer, **kwargs):
        for name in trainer.model_names:
            super(GradientVisualize, self).__call__(name, **kwargs)
            self.logs[name].append(0.0)


class ImageVisualize(Visualize):
    r"""This class provides the Logging for the Images.

    Args:
        trainer (torchgan.trainer.Trainer): The base trainer used for training.
        visdom_port (int, optional): Port to log using ``visdom``. The visdom server needs to be
            manually started at this port else an error will be thrown and the code will crash.
            This is ignored if ``VISDOM_LOGGING`` is ``0``.
        log_dir (str, optional): Directory where TensorboardX should store the logs. This is
            ignored if ``TENSORBOARD_LOGGING`` is ``0``.
        writer (tensorboardX.SummaryWriter, optonal): Send a `SummaryWriter` if you
            don't want to start a new SummaryWriter.
        test_noise (torch.Tensor, optional): If provided then it will be used as the noise for image
            sampling.
        nrow (int, optional): Number of rows in which the image is to be stored.
    """

    def __init__(
        self,
        trainer,
        visdom_port=8097,
        log_dir=None,
        writer=None,
        test_noise=None,
        nrow=8,
    ):
        super(ImageVisualize, self).__init__(
            [], visdom_port=visdom_port, log_dir=log_dir, writer=writer
        )
        self.test_noise = []
        for model in trainer.model_names:
            if isinstance(getattr(trainer, model), Generator):
                self.test_noise.append(
                    getattr(trainer, model).sampler(trainer.sample_size, trainer.device)
                    if test_noise is None
                    else test_noise
                )
        self.step = 1
        self.nrow = nrow

    def log_tensorboard(self, trainer, image, model):
        r"""Logs a generated image in tensorboard at the end of an epoch.

        Args:
            trainer (torchgan.trainer.Trainer): The base trainer used for training.
            image (Image): The generated image.
            model (str): The name of the model which generated the ``image``.
        """
        self.writer.add_image("Generated Samples/{}".format(model), image, self.step)

    def log_console(self, trainer, image, model):
        r"""Saves a generated image at the end of an epoch. The path where the image is
        being stored is controlled by the ``trainer``.

        Args:
            trainer (torchgan.trainer.Trainer): The base trainer used for training.
            image (Image): The generated image.
            model (str): The name of the model which generated the ``image``.
        """
        save_path = "{}/epoch{}_{}.png".format(trainer.recon, self.step, model)
        print("Generating and Saving Images to {}".format(save_path))
        torchvision.utils.save_image(image, save_path)

    def log_visdom(self, trainer, image, model):
        r"""Logs a generated image in visdom at the end of an epoch.

        Args:
            trainer (torchgan.trainer.Trainer): The base trainer used for training.
            image (Image): The generated image.
            model (str): The name of the model which generated the ``image``.
        """
        self.vis.image(image, opts=dict(caption="Generated Samples/{}".format(model)))

    def __call__(self, trainer, **kwargs):
        pos = 0
        for model in trainer.model_names:
            if isinstance(getattr(trainer, model), Generator):
                generator = getattr(trainer, model)
                with torch.no_grad():
                    image = generator(*self.test_noise[pos])
                    image = torchvision.utils.make_grid(
                        image, nrow=self.nrow, normalize=True, range=(-1, 1)
                    )
                    super(ImageVisualize, self).__call__(
                        trainer, image, model, **kwargs
                    )
                self.step -= 1
                pos = pos + 1
        self.step += 1 if pos > 0 else 0



import torch

from .functional import minimax_discriminator_loss, minimax_generator_loss
from .loss import DiscriminatorLoss, GeneratorLoss

__all__ = ["MinimaxGeneratorLoss", "MinimaxDiscriminatorLoss"]


class MinimaxGeneratorLoss(GeneratorLoss):
    r"""Minimax game generator loss from the original GAN paper `"Generative Adversarial Networks
    by Goodfellow et. al." <https://arxiv.org/abs/1406.2661>`_

    The loss can be described as:

    .. math:: L(G) = log(1 - D(G(z)))

    The nonsaturating heuristic is also supported:

    .. math:: L(G) = -log(D(G(z)))

    where

    - :math:`G` : Generator
    - :math:`D` : Discriminator
    - :math:`z` : A sample from the noise prior

    Args:
        reduction (str, optional): Specifies the reduction to apply to the output.
            If ``none`` no reduction will be applied. If ``mean`` the outputs are averaged over batch size.
            If ``sum`` the elements of the output are summed.
        override_train_ops (function, optional): Function to be used in place of the default ``train_ops``
        nonsaturating(bool, optional): Specifies whether to use the nonsaturating heuristic
            loss for the generator.
    """

    def __init__(self, reduction="mean", nonsaturating=True, override_train_ops=None):
        super(MinimaxGeneratorLoss, self).__init__(reduction, override_train_ops)
        self.nonsaturating = nonsaturating

    def forward(self, dgz):
        r"""Computes the loss for the given input.

        Args:
            dgz (torch.Tensor) : Output of the Discriminator with generated data. It must have the
                                 dimensions (N, \*) where \* means any number of additional
                                 dimensions.

        Returns:
            scalar if reduction is applied else Tensor with dimensions (N, \*).
        """
        return minimax_generator_loss(dgz, self.nonsaturating, self.reduction)


class MinimaxDiscriminatorLoss(DiscriminatorLoss):
    r"""Minimax game discriminator loss from the original GAN paper `"Generative Adversarial Networks
    by Goodfellow et. al." <https://arxiv.org/abs/1406.2661>`_

    The loss can be described as:

    .. math:: L(D) = -[log(D(x)) + log(1 - D(G(z)))]

    where

    - :math:`G` : Generator
    - :math:`D` : Discriminator
    - :math:`x` : A sample from the data distribution
    - :math:`z` : A sample from the noise prior

    Args:
        label_smoothing (float, optional): The factor by which the labels (1 in this case) needs
            to be smoothened. For example, label_smoothing = 0.2 changes the value of the real
            labels to 0.8.
        reduction (str, optional): Specifies the reduction to apply to the output.
            If ``none`` no reduction will be applied. If ``mean`` the mean of the output.
            If ``sum`` the elements of the output will be summed.
        override_train_ops (function, optional): A function is passed to this argument,
            if the default ``train_ops`` is not to be used.
    """

    def __init__(self, label_smoothing=0.0, reduction="mean", override_train_ops=None):
        super(MinimaxDiscriminatorLoss, self).__init__(reduction, override_train_ops)
        self.label_smoothing = label_smoothing

    def forward(self, dx, dgz):
        r"""Computes the loss for the given input.

        Args:
            dx (torch.Tensor) : Output of the Discriminator with real data. It must have the
                                dimensions (N, \*) where \* means any number of additional
                                dimensions.
            dgz (torch.Tensor) : Output of the Discriminator with generated data. It must have the
                                 dimensions (N, \*) where \* means any number of additional
                                 dimensions.

        Returns:
            scalar if reduction is applied else Tensor with dimensions (N, \*).
        """
        return minimax_discriminator_loss(
            dx, dgz, label_smoothing=self.label_smoothing, reduction=self.reduction
        )



import os
import sys
import unittest

import torch
import torch.utils.data as data
import torchvision
import torchvision.datasets as dsets
import torchvision.transforms as transforms
from torch.optim import Adam
from torchgan import *
from torchgan.losses import *
from torchgan.metrics import *
from torchgan.models import *
from torchgan.trainer import Trainer

sys.path.append(os.path.join(os.path.dirname(__file__), "../.."))


def mnist_dataloader():
    train_dataset = dsets.MNIST(
        root="./mnist",
        train=True,
        transform=transforms.Compose(
            [
                transforms.Pad((2, 2)),
                transforms.ToTensor(),
                transforms.Normalize(mean=(0.5,), std=(0.5,)),
            ]
        ),
        download=True,
    )
    train_loader = data.DataLoader(train_dataset, batch_size=128, shuffle=True)
    return train_loader


class TestTrainer(unittest.TestCase):
    def test_trainer_dcgan(self):
        network_params = {
            "generator": {
                "name": DCGANGenerator,
                "args": {"out_channels": 1, "step_channels": 4},
                "optimizer": {
                    "name": Adam,
                    "args": {"lr": 0.0002, "betas": (0.5, 0.999)},
                },
            },
            "discriminator": {
                "name": DCGANDiscriminator,
                "args": {"in_channels": 1, "step_channels": 4},
                "optimizer": {
                    "name": Adam,
                    "args": {"lr": 0.0002, "betas": (0.5, 0.999)},
                },
            },
        }
        losses_list = [MinimaxGeneratorLoss(), MinimaxDiscriminatorLoss()]
        trainer = Trainer(
            network_params,
            losses_list,
            sample_size=1,
            epochs=1,
            device=torch.device("cpu"),
        )
        trainer(mnist_dataloader())

    def test_trainer_cgan(self):
        network_params = {
            "generator": {
                "name": ConditionalGANGenerator,
                "args": {"num_classes": 10, "out_channels": 1, "step_channels": 4},
                "optimizer": {
                    "name": Adam,
                    "args": {"lr": 0.0002, "betas": (0.5, 0.999)},
                },
            },
            "discriminator": {
                "name": ConditionalGANDiscriminator,
                "args": {"num_classes": 10, "in_channels": 1, "step_channels": 4},
                "optimizer": {
                    "name": Adam,
                    "args": {"lr": 0.0002, "betas": (0.5, 0.999)},
                },
            },
        }
        losses_list = [MinimaxGeneratorLoss(), MinimaxDiscriminatorLoss()]
        trainer = Trainer(
            network_params,
            losses_list,
            sample_size=1,
            epochs=1,
            device=torch.device("cpu"),
        )
        trainer(mnist_dataloader())

    def test_trainer_acgan(self):
        network_params = {
            "generator": {
                "name": ACGANGenerator,
                "args": {"num_classes": 10, "out_channels": 1, "step_channels": 4},
                "optimizer": {
                    "name": Adam,
                    "args": {"lr": 0.0002, "betas": (0.5, 0.999)},
                },
            },
            "discriminator": {
                "name": ACGANDiscriminator,
                "args": {"num_classes": 10, "in_channels": 1, "step_channels": 4},
                "optimizer": {
                    "name": Adam,
                    "args": {"lr": 0.0002, "betas": (0.5, 0.999)},
                },
            },
        }
        losses_list = [
            MinimaxGeneratorLoss(),
            MinimaxDiscriminatorLoss(),
            AuxiliaryClassifierGeneratorLoss(),
            AuxiliaryClassifierDiscriminatorLoss(),
        ]
        trainer = Trainer(
            network_params,
            losses_list,
            sample_size=1,
            epochs=1,
            device=torch.device("cpu"),
        )
        trainer(mnist_dataloader())



import torch

from .functional import dragan_gradient_penalty
from .loss import DiscriminatorLoss, GeneratorLoss

__all__ = ["DraganGradientPenalty"]


class DraganGradientPenalty(DiscriminatorLoss):
    r"""Gradient Penalty for the DRAGAN discriminator from `"On Convergence and Stability of GANs
    by Kodali et. al." <https://arxiv.org/abs/1705.07215>`_ paper

    The gradient penalty is calculated as:

    .. math:: \lambda \times (||grad(D(x))||_2 - k)^2

    The gradient being taken with respect to x

    where

    - :math:`G` : Generator
    - :math:`D` : Disrciminator
    - :math:`\lambda` : Scaling hyperparameter
    - :math:`x` : Interpolation term for the gradient penalty
    - :math:`k` : Constant

    Args:
        reduction (str, optional): Specifies the reduction to apply to the output.
            If ``none`` no reduction will be applied. If ``mean`` the outputs are averaged over batch size.
            If ``sum`` the elements of the output are summed.
        lambd (float,optional) : Hyperparameter :math:`\lambda` for scaling the gradient penalty.
        k (float, optional) : Constant.
        override_train_ops (function, optional): Function to be used in place of the default ``train_ops``
    """

    def __init__(self, reduction="mean", lambd=10.0, k=1.0, override_train_ops=None):
        super(DraganGradientPenalty, self).__init__(reduction)
        self.lambd = lambd
        self.override_train_ops = override_train_ops
        self.k = k

    def forward(self, interpolate, d_interpolate):
        r"""Computes the loss for the given input.

        Args:
            interpolate (torch.Tensor) : It must have the dimensions (N, \*) where
                                         \* means any number of additional dimensions.
            d_interpolate (torch.Tensor) : Output of the ``discriminator`` with ``interpolate``
                                           as the input. It must have the dimensions (N, \*)
                                           where \* means any number of additional dimensions.

        Returns:
            scalar if reduction is applied else Tensor with dimensions (N, \*).
        """
        return dragan_gradient_penalty(
            interpolate, d_interpolate, self.k, self.reduction
        )

    def train_ops(
        self,
        generator,
        discriminator,
        optimizer_discriminator,
        real_inputs,
        device,
        labels=None,
    ):
        r"""Defines the standard ``train_ops`` used by the DRAGAN Gradient Penalty.

        The ``standard optimization algorithm`` for the ``discriminator`` defined in this train_ops
        is as follows:

        1. :math:`interpolate = real + \frac{1}{2} \times (1 - \alpha) \times std(real) \times \beta`
        2. :math:`d\_interpolate = discriminator(interpolate)`
        3. :math:`loss = loss\_function(interpolate, d\_interpolate)`
        4. Backpropagate by computing :math:`\nabla loss`
        5. Run a step of the optimizer for discriminator

        Args:
            generator (torchgan.models.Generator): The model to be optimized.
            discriminator (torchgan.models.Discriminator): The discriminator which judges the
                performance of the generator.
            optimizer_discriminator (torch.optim.Optimizer): Optimizer which updates the ``parameters``
                of the ``discriminator``.
            real_inputs (torch.Tensor): The real data to be fed to the ``discriminator``.
            device (torch.device): Device on which the ``generator`` and ``discriminator`` is present.
            labels (torch.Tensor, optional): Labels for the data.

        Returns:
            Scalar value of the loss.
        """
        if self.override_train_ops is not None:
            return self.override_train_ops(
                self,
                generator,
                discriminator,
                optimizer_discriminator,
                real_inputs,
                labels,
            )
        else:
            # NOTE(avik-pal): We don't need the gradients for alpha and beta. It's there
            #                 to prevent an error while calling autograd.grad
            alpha = torch.rand(
                size=real_inputs.shape, device=device, requires_grad=True
            )
            beta = torch.rand(size=real_inputs.shape, device=device, requires_grad=True)
            optimizer_discriminator.zero_grad()
            interpolate = real_inputs + (1 - alpha) * 0.5 * real_inputs.std() * beta
            if generator.label_type == "generated":
                label_gen = torch.randint(
                    0, generator.num_classes, (real_inputs.size(0),), device=device
                )
            if discriminator.label_type == "none":
                d_interpolate = discriminator(interpolate)
            else:
                if generator.label_type == "generated":
                    d_interpolate = discriminator(interpolate, label_gen)
                else:
                    d_interpolate = discriminator(interpolate, labels)
            loss = self.forward(interpolate, d_interpolate)
            weighted_loss = self.lambd * loss
            weighted_loss.backward()
            optimizer_discriminator.step()
            return loss.item()



import os
import time
from inspect import _empty, signature
from warnings import warn

import torch
import torchvision

from ..logging.logger import Logger
from ..losses.loss import DiscriminatorLoss, GeneratorLoss
from ..models.model import Discriminator, Generator

__all__ = ["BaseTrainer"]


class BaseTrainer(object):
    r"""Base Trainer for TorchGANs.

    .. warning::
        This trainer is meant to form the base for all other Trainers. This is not meant for direct usage.

    Features provided by this Base Trainer are:

    - Loss and Metrics Logging via the ``Logger`` class.
    - Generating Image Samples.
    - Saving models at the end of every epoch and loading of previously saved models.
    - Highly flexible and allows changing hyperparameters by simply adjusting the arguments.

    Most of the functionalities provided by the Trainer are flexible enough and can be customized by
    simply passing different arguments. You can train anything from a simple DCGAN to complex CycleGANs
    without ever having to subclass this ``Trainer``.

    Args:
        losses_list (list): A list of the Loss Functions that need to be minimized. For a list of
            pre-defined losses look at :mod:`torchgan.losses`. All losses in the list must be a
            subclass of atleast ``GeneratorLoss`` or ``DiscriminatorLoss``.
        metrics_list (list, optional): List of Metric Functions that need to be logged. For a list of
            pre-defined metrics look at :mod:`torchgan.metrics`. All losses in the list must be a
            subclass of ``EvaluationMetric``.
        device (torch.device, optional): Device in which the operation is to be carried out. If you
            are using a CPU machine make sure that you change it for proper functioning.
        ncritic (int, optional): Setting it to a value will make the discriminator train that many
            times more than the generator. If it is set to a negative value the generator will be
            trained that many times more than the discriminator.
        sample_size (int, optional): Total number of images to be generated at the end of an epoch
            for logging purposes.
        epochs (int, optional): Total number of epochs for which the models are to be trained.
        checkpoints (str, optional): Path where the models are to be saved. The naming convention is
            if checkpoints is ``./model/gan`` then models are saved as ``./model/gan0.model`` and so on.
        retain_checkpoints (int, optional): Total number of checkpoints that should be retained. For
            example, if the value is set to 3, we save at most 3 models and start rewriting the models
            after that.
        recon (str, optional): Directory where the sampled images are saved. Make sure the directory
            exists from beforehand.
        log_dir (str, optional): The directory for logging tensorboard. It is ignored if
            TENSORBOARD_LOGGING is 0.
        test_noise (torch.Tensor, optional): If provided then it will be used as the noise for image
            sampling.
        nrow (int, optional): Number of rows in which the image is to be stored.

    Any other argument that you need to store in the object can be simply passed via keyword arguments.
    """

    def __init__(
        self,
        losses_list,
        metrics_list=None,
        device=torch.device("cuda:0"),
        ncritic=1,
        epochs=5,
        sample_size=8,
        checkpoints="./model/gan",
        retain_checkpoints=5,
        recon="./images",
        log_dir=None,
        test_noise=None,
        nrow=8,
        **kwargs,
    ):
        self.device = device
        self.losses = {}
        for loss in losses_list:
            self.losses[type(loss).__name__] = loss

        if metrics_list is None:
            self.metrics = None
        else:
            self.metrics = {}
            for metric in metrics_list:
                self.metrics[type(metric).__name__] = metric

        self.sample_size = sample_size
        self.epochs = epochs
        self.checkpoints = checkpoints
        self.retain_checkpoints = retain_checkpoints
        self.recon = recon

        # Not needed but we need to store this to avoid errors. Also makes life simpler
        self.noise = None
        self.real_inputs = None
        self.labels = None
        self.batch_size = 1

        self.loss_information = {
            "generator_losses": 0.0,
            "discriminator_losses": 0.0,
            "generator_iters": 0,
            "discriminator_iters": 0,
        }

        assert ncritic != 0
        if ncritic > 0:
            self.ncritic = ncritic
            self.ngen = 1
        else:
            self.ncritic = 1
            self.ngen = abs(ncritic)

        self.start_epoch = 0
        self.last_retained_checkpoint = 0
        for key, val in kwargs.items():
            if key in self.__dict__:
                warn(
                    "Overiding the default value of {} from {} to {}".format(
                        key, getattr(self, key), val
                    )
                )
            setattr(self, key, val)

        os.makedirs(self.checkpoints.rsplit("/", 1)[0], exist_ok=True)
        os.makedirs(self.recon, exist_ok=True)

    def save_model(self, epoch, save_items=None):
        r"""Function saves the model and some necessary information along with it. List of items
        stored for future reference:

        - Epoch
        - Model States
        - Optimizer States
        - Loss Information
        - Loss Objects
        - Metric Objects
        - Loss Logs

        The save location is printed when this function is called.

        Args:
            epoch (int, optional): Epoch Number at which the model is being saved
            save_items (str, list, optional): Pass the variable name of any other item you want to save.
                The item must be present in the `__dict__` else training will come to an abrupt end.
        """
        if self.last_retained_checkpoint == self.retain_checkpoints:
            self.last_retained_checkpoint = 0
        save_path = self.checkpoints + str(self.last_retained_checkpoint) + ".model"
        self.last_retained_checkpoint += 1
        print("Saving Model at '{}'".format(save_path))
        model = {
            "epoch": epoch + 1,
            "loss_information": self.loss_information,
            "loss_objects": self.losses,
            "metric_objects": self.metrics,
            "loss_logs": (self.logger.get_loss_viz()).logs,
        }
        for save_item in self.model_names + self.optimizer_names:
            model.update({save_item: (getattr(self, save_item)).state_dict()})
        if save_items is not None:
            if type(save_items) is list:
                for itms in save_items:
                    model.update({itms: getattr(self, itms)})
            else:
                model.update({save_items: getattr(self, save_items)})
        torch.save(model, save_path)

    def load_model(self, load_path="", load_items=None):
        r"""Function to load the model and some necessary information along with it. List of items
        loaded:

        - Epoch
        - Model States
        - Optimizer States
        - Loss Information
        - Loss Objects
        - Metric Objects
        - Loss Logs

        .. warning::
            An Exception is raised if the model could not be loaded. Make sure that the model being
            loaded was saved previously by ``torchgan Trainer`` itself. We currently do not support
            loading any other form of models but this might be improved in the future.

        Args:
            load_path (str, optional): Path from which the model is to be loaded.
            load_items (str, list, optional): Pass the variable name of any other item you want to load.
                If the item cannot be found then a warning will be thrown and model will start to train
                from scratch. So make sure that item was saved.
        """
        if load_path == "":
            load_path = self.checkpoints + str(self.last_retained_checkpoint) + ".model"
        print("Loading Model From '{}'".format(load_path))
        try:
            checkpoint = torch.load(load_path)
            self.start_epoch = checkpoint["epoch"]
            self.losses = checkpoint["loss_objects"]
            self.metrics = checkpoint["metric_objects"]
            self.loss_information = checkpoint["loss_information"]
            (self.logger.get_loss_viz()).logs = checkpoint["loss_logs"]
            for load_item in self.model_names + self.optimizer_names:
                getattr(self, load_item).load_state_dict(checkpoint[load_item])
            if load_items is not None:
                if type(load_items) is list:
                    for itms in load_items:
                        setattr(self, itms, checkpoint["itms"])
                else:
                    setattr(self, load_items, checkpoint["load_items"])
        except:
            raise Exception("Model could not be loaded from {}.".format(load_path))

    def _get_argument_maps(self, default_map, func):
        r"""Extracts the signature of the `func`. Then it returns the list of arguments that
        are present in the object and need to be mapped and passed to the `func` when calling it.

        Args:
            default_map (dict): The keys of this dictionary override the function signature.
            func (function): Function whose argument map is to be generated.

        Returns:
            List of arguments that need to be fed into the function. It contains all the positional
            arguments and keyword arguments that are stored in the object. If any of the required
            arguments are not present an error is thrown.
        """
        sig = signature(func)
        arg_map = {}
        for sig_param in sig.parameters.values():
            arg = sig_param.name
            arg_name = arg
            if arg in default_map:
                arg_name = default_map[arg]
            if sig_param.default is not _empty:
                if arg_name in self.__dict__:
                    arg_map.update({arg: arg_name})
            else:
                if arg_name not in self.__dict__ and arg != "kwargs" and arg != "args":
                    raise Exception("Argument : {} not present.".format(arg_name))
                else:
                    arg_map.update({arg: arg_name})
        return arg_map

    def _store_metric_maps(self):
        r"""Creates a mapping between the metrics and the arguments from the object that need to be
        passed to it.
        """
        if self.metrics is not None:
            self.metric_arg_maps = {}
            for name, metric in self.metrics.items():
                self.metric_arg_maps[name] = self._get_argument_maps(
                    metric.arg_map, metric.metric_ops
                )

    def _store_loss_maps(self):
        r"""Creates a mapping between the losses and the arguments from the object that need to be
        passed to it.
        """
        self.loss_arg_maps = {}
        for name, loss in self.losses.items():
            self.loss_arg_maps[name] = self._get_argument_maps(
                loss.arg_map, loss.train_ops
            )

    def _get_arguments(self, arg_map):
        r"""Get the argument values from the object and create a dictionary.

        Args:
            arg_map (dict): A dict of arguments that is generated by `_get_argument_maps`.

        Returns:
            A dictionary mapping the argument name to the value of the argument.
        """
        args = {}
        for key, val in arg_map.items():
            args[key] = self.__dict__[val]
        return args

    def train_iter_custom(self):
        r"""Function that needs to be extended if ``train_iter`` is to be modified. Use this function
        to perform any sort of initialization that need to be done at the beginning of any train
        iteration. Refer the model zoo and tutorials for more details on how to write this function.
        """
        pass

    # TODO(avik-pal): Clean up this function and avoid returning values
    def train_iter(self):
        r"""Calls the train_ops of the loss functions. This is the core function of the Trainer. In most
        cases you will never have the need to extend this function. In extreme cases simply extend
        ``train_iter_custom``.

        .. warning::
            This function is needed in this exact state for the Trainer to work correctly. So it is
            highly recommended that this function is not changed even if the ``Trainer`` is subclassed.

        Returns:
            An NTuple of the ``generator loss``, ``discriminator loss``, ``number of times the generator
            was trained`` and the ``number of times the discriminator was trained``.
        """
        self.train_iter_custom()
        ldis, lgen, dis_iter, gen_iter = 0.0, 0.0, 0, 0
        loss_logs = self.logger.get_loss_viz()
        grad_logs = self.logger.get_grad_viz()
        for name, loss in self.losses.items():
            if isinstance(loss, GeneratorLoss) and isinstance(loss, DiscriminatorLoss):
                # NOTE(avik-pal): In most cases this loss is meant to optimize the Discriminator
                #                 but we might need to think of a better solution
                if self.loss_information["generator_iters"] % self.ngen == 0:
                    cur_loss = loss.train_ops(
                        **self._get_arguments(self.loss_arg_maps[name])
                    )
                    loss_logs.logs[name].append(cur_loss)
                    if type(cur_loss) is tuple:
                        lgen, ldis, gen_iter, dis_iter = (
                            lgen + cur_loss[0],
                            ldis + cur_loss[1],
                            gen_iter + 1,
                            dis_iter + 1,
                        )
                    else:
                        # NOTE(avik-pal): We assume that it is a Discriminator Loss by default.
                        ldis, dis_iter = ldis + cur_loss, dis_iter + 1
                for model_name in self.model_names:
                    grad_logs.update_grads(model_name, getattr(self, model_name))
            elif isinstance(loss, GeneratorLoss):
                if self.loss_information["discriminator_iters"] % self.ncritic == 0:
                    cur_loss = loss.train_ops(
                        **self._get_arguments(self.loss_arg_maps[name])
                    )
                    loss_logs.logs[name].append(cur_loss)
                    lgen, gen_iter = lgen + cur_loss, gen_iter + 1
                for model_name in self.model_names:
                    model = getattr(self, model_name)
                    if isinstance(model, Generator):
                        grad_logs.update_grads(model_name, model)
            elif isinstance(loss, DiscriminatorLoss):
                if self.loss_information["generator_iters"] % self.ngen == 0:
                    cur_loss = loss.train_ops(
                        **self._get_arguments(self.loss_arg_maps[name])
                    )
                    loss_logs.logs[name].append(cur_loss)
                    ldis, dis_iter = ldis + cur_loss, dis_iter + 1
                for model_name in self.model_names:
                    model = getattr(self, model_name)
                    if isinstance(model, Discriminator):
                        grad_logs.update_grads(model_name, model)
        return lgen, ldis, gen_iter, dis_iter

    def eval_ops(self, **kwargs):
        r"""Runs all evaluation operations at the end of every epoch. It calls all the metric functions
        that are passed to the Trainer.
        """
        if self.metrics is not None:
            for name, metric in self.metrics.items():
                metric_logs = self.logger.get_metric_viz()
                metric_logs.logs[name].append(
                    metric.metric_ops(**self._get_arguments(self.metric_arg_maps[name]))
                )

    def optim_ops(self):
        r"""Runs all the schedulers at the end of every epoch.
        """
        for scheduler in self.schedulers:
            scheduler.step()

    def train(self, data_loader, **kwargs):
        r"""Uses the information passed by the user while creating the object and trains the model.
        It iterates over the epochs and the DataLoader and calls the functions for training the models
        and logging the required variables.

        .. note::
            Even though ``__call__`` calls this function, it is best if ``train`` is not called directly.
            When ``__call__`` is invoked, we infer the ``batch_size`` from the ``data_loader``. Also,
            we are certain not going to change the interface of the ``__call__`` function so it gives
            the user a stable API, while we can change the flow of execution of ``train`` in future.

        .. warning::
            The user should never try to change this function in subclass. It is too delicate and
            changing affects every other function present in this ``Trainer`` class.

        This function controls the execution of all the components of the ``Trainer``. It controls the
        ``logger``, ``train_iter``, ``save_model``, ``eval_ops`` and ``optim_ops``.

        Args:
            data_loader (torch.utils.data.DataLoader): A DataLoader for the trainer to iterate over
                and train the models.
        """
        for name in self.optimizer_names:
            getattr(self, name).zero_grad()

        for epoch in range(self.start_epoch, self.epochs):

            start_time = time.time()

            for model in self.model_names:
                getattr(self, model).train()

            for data in data_loader:

                if type(data) is tuple or type(data) is list:
                    self.real_inputs = data[0].to(self.device)
                    self.labels = data[1].to(self.device)
                elif type(data) is torch.Tensor:
                    self.real_inputs = data.to(self.device)
                else:
                    self.real_inputs = data

                lgen, ldis, gen_iter, dis_iter = self.train_iter()
                self.loss_information["generator_losses"] += lgen
                self.loss_information["discriminator_losses"] += ldis
                self.loss_information["generator_iters"] += gen_iter
                self.loss_information["discriminator_iters"] += dis_iter

                self.logger.run_mid_epoch(self)

            if "save_items" in kwargs:
                self.save_model(epoch, kwargs["save_items"])
            else:
                self.save_model(epoch)

            for model in self.model_names:
                getattr(self, model).eval()

            self.eval_ops(**kwargs)
            self.logger.run_end_epoch(self, epoch, time.time() - start_time)
            self.optim_ops()

        print("Training of the Model is Complete")

    def complete(self, **kwargs):
        r"""Marks the end of training. It saves the final model and turns off the logger.

        .. note::
            It is not necessary to call this function. If it is not called the logger is kept
            alive in the background. So it might be considered a good practice to call this
            function.
        """
        if "save_items" in kwargs:
            self.save_model(-1, kwargs["save_items"])
        else:
            self.save_model(-1)
        self.logger.close()

    def __call__(self, data_loader, **kwargs):
        self.batch_size = data_loader.batch_size
        self.train(data_loader, **kwargs)



import torch

from .functional import (
    wasserstein_discriminator_loss,
    wasserstein_generator_loss,
    wasserstein_gradient_penalty,
)
from .loss import DiscriminatorLoss, GeneratorLoss

__all__ = [
    "WassersteinGeneratorLoss",
    "WassersteinDiscriminatorLoss",
    "WassersteinGradientPenalty",
]


class WassersteinGeneratorLoss(GeneratorLoss):
    r"""Wasserstein GAN generator loss from
    `"Wasserstein GAN by Arjovsky et. al." <https://arxiv.org/abs/1701.07875>`_ paper

    The loss can be described as:

    .. math:: L(G) = -f(G(z))

    where

    - :math:`G` : Generator
    - :math:`f` : Critic/Discriminator
    - :math:`z` : A sample from the noise prior

    Args:
        reduction (str, optional): Specifies the reduction to apply to the output.
            If ``none`` no reduction will be applied. If ``mean`` the mean of the output.
            If ``sum`` the elements of the output will be summed.
        override_train_ops (function, optional): A function is passed to this argument,
            if the default ``train_ops`` is not to be used.
    """

    def forward(self, fgz):
        r"""Computes the loss for the given input.

        Args:
            dgz (torch.Tensor) : Output of the Discriminator with generated data. It must have the
                                 dimensions (N, \*) where \* means any number of additional
                                 dimensions.

        Returns:
            scalar if reduction is applied else Tensor with dimensions (N, \*).
        """
        return wasserstein_generator_loss(fgz, self.reduction)


class WassersteinDiscriminatorLoss(DiscriminatorLoss):
    r"""Wasserstein GAN generator loss from
    `"Wasserstein GAN by Arjovsky et. al." <https://arxiv.org/abs/1701.07875>`_ paper

    The loss can be described as:

    .. math:: L(D) = f(G(z)) - f(x)

    where

    - :math:`G` : Generator
    - :math:`f` : Critic/Discriminator
    - :math:`x` : A sample from the data distribution
    - :math:`z` : A sample from the noise prior

    Args:
        reduction (str, optional): Specifies the reduction to apply to the output.
            If ``none`` no reduction will be applied. If ``mean`` the mean of the output.
            If ``sum`` the elements of the output will be summed.
        clip (tuple, optional): Tuple that specifies the maximum and minimum parameter
            clamping to be applied, as per the original version of the Wasserstein loss
            without Gradient Penalty.
        override_train_ops (function, optional): A function is passed to this argument,
            if the default ``train_ops`` is not to be used.
    """

    def __init__(self, reduction="mean", clip=None, override_train_ops=None):
        super(WassersteinDiscriminatorLoss, self).__init__(
            reduction, override_train_ops
        )
        if (isinstance(clip, tuple) or isinstance(clip, list)) and len(clip) > 1:
            self.clip = clip
        else:
            self.clip = None

    def forward(self, fx, fgz):
        r"""Computes the loss for the given input.

        Args:
            fx (torch.Tensor) : Output of the Discriminator with real data. It must have the
                                dimensions (N, \*) where \* means any number of additional
                                dimensions.
            fgz (torch.Tensor) : Output of the Discriminator with generated data. It must have the
                                 dimensions (N, \*) where \* means any number of additional
                                 dimensions.

        Returns:
            scalar if reduction is applied else Tensor with dimensions (N, \*).
        """
        return wasserstein_discriminator_loss(fx, fgz, self.reduction)

    def train_ops(
        self,
        generator,
        discriminator,
        optimizer_discriminator,
        real_inputs,
        device,
        labels=None,
    ):
        r"""Defines the standard ``train_ops`` used by wasserstein discriminator loss.

        The ``standard optimization algorithm`` for the ``discriminator`` defined in this train_ops
        is as follows:

        1. Clamp the discriminator parameters to satisfy :math:`lipschitz\ condition`
        2. :math:`fake = generator(noise)`
        3. :math:`value_1 = discriminator(fake)`
        4. :math:`value_2 = discriminator(real)`
        5. :math:`loss = loss\_function(value_1, value_2)`
        6. Backpropagate by computing :math:`\nabla loss`
        7. Run a step of the optimizer for discriminator

        Args:
            generator (torchgan.models.Generator): The model to be optimized.
            discriminator (torchgan.models.Discriminator): The discriminator which judges the
                performance of the generator.
            optimizer_discriminator (torch.optim.Optimizer): Optimizer which updates the ``parameters``
                of the ``discriminator``.
            real_inputs (torch.Tensor): The real data to be fed to the ``discriminator``.
            device (torch.device): Device on which the ``generator`` and ``discriminator`` is present.
            labels (torch.Tensor, optional): Labels for the data.

        Returns:
            Scalar value of the loss.
        """
        if self.override_train_ops is not None:
            return self.override_train_ops(
                generator,
                discriminator,
                optimizer_discriminator,
                real_inputs,
                device,
                labels,
            )
        else:
            if self.clip is not None:
                for p in discriminator.parameters():
                    p.data.clamp_(self.clip[0], self.clip[1])
            return super(WassersteinDiscriminatorLoss, self).train_ops(
                generator,
                discriminator,
                optimizer_discriminator,
                real_inputs,
                device,
                labels,
            )


class WassersteinGradientPenalty(DiscriminatorLoss):
    r"""Gradient Penalty for the Improved Wasserstein GAN discriminator from
    `"Improved Training of Wasserstein GANs
    by Gulrajani et. al." <https://arxiv.org/abs/1704.00028>`_ paper

    The gradient penalty is calculated as:

    .. math: \lambda \times (||\nabla(D(x))||_2 - 1)^2

    The gradient being taken with respect to x

    where

    - :math:`G` : Generator
    - :math:`D` : Disrciminator/Critic
    - :math:`\lambda` : Scaling hyperparameter
    - :math:`x` : Interpolation term for the gradient penalty

    Args:
        reduction (str, optional): Specifies the reduction to apply to the output.
            If ``none`` no reduction will be applied. If ``mean`` the mean of the output.
            If ``sum`` the elements of the output will be summed.
        lambd (float,optional): Hyperparameter lambda for scaling the gradient penalty.
        override_train_ops (function, optional): A function is passed to this argument,
            if the default ``train_ops`` is not to be used.
    """

    def __init__(self, reduction="mean", lambd=10.0, override_train_ops=None):
        super(WassersteinGradientPenalty, self).__init__(reduction, override_train_ops)
        self.lambd = lambd
        self.override_train_ops = override_train_ops

    def forward(self, interpolate, d_interpolate):
        r"""Computes the loss for the given input.

        Args:
            interpolate (torch.Tensor) : It must have the dimensions (N, \*) where
                                         \* means any number of additional dimensions.
            d_interpolate (torch.Tensor) : Output of the ``discriminator`` with ``interpolate``
                                           as the input. It must have the dimensions (N, \*)
                                           where \* means any number of additional dimensions.

        Returns:
            scalar if reduction is applied else Tensor with dimensions (N, \*).
        """
        # TODO(Aniket1998): Check for performance bottlenecks
        # If found, write the backprop yourself instead of
        # relying on autograd
        return wasserstein_gradient_penalty(interpolate, d_interpolate, self.reduction)

    def train_ops(
        self,
        generator,
        discriminator,
        optimizer_discriminator,
        real_inputs,
        device,
        labels=None,
    ):
        r"""Defines the standard ``train_ops`` used by the Wasserstein Gradient Penalty.

        The ``standard optimization algorithm`` for the ``discriminator`` defined in this train_ops
        is as follows:

        1. :math:`fake = generator(noise)`
        2. :math:`interpolate = \epsilon \times real + (1 - \epsilon) \times fake`
        3. :math:`d\_interpolate = discriminator(interpolate)`
        4. :math:`loss = \lambda loss\_function(interpolate, d\_interpolate)`
        5. Backpropagate by computing :math:`\nabla loss`
        6. Run a step of the optimizer for discriminator

        Args:
            generator (torchgan.models.Generator): The model to be optimized.
            discriminator (torchgan.models.Discriminator): The discriminator which judges the
                performance of the generator.
            optimizer_discriminator (torch.optim.Optimizer): Optimizer which updates the ``parameters``
                of the ``discriminator``.
            real_inputs (torch.Tensor): The real data to be fed to the ``discriminator``.
            device (torch.device): Device on which the ``generator`` and ``discriminator`` is present.
            batch_size (int): Batch Size of the data infered from the ``DataLoader`` by the ``Trainer``.
            labels (torch.Tensor, optional): Labels for the data.

        Returns:
            Scalar value of the loss.
        """
        if self.override_train_ops is not None:
            return self.override_train_ops(
                self,
                generator,
                discriminator,
                optimizer_discriminator,
                real_inputs,
                labels,
            )
        else:
            if labels is None and (
                generator.label_type == "required"
                or discriminator.label_type == "required"
            ):
                raise Exception("GAN model requires labels for training")
            batch_size = real_inputs.size(0)
            noise = torch.randn(batch_size, generator.encoding_dims, device=device)
            if generator.label_type == "generated":
                label_gen = torch.randint(
                    0, generator.num_classes, (batch_size,), device=device
                )
            optimizer_discriminator.zero_grad()
            if generator.label_type == "none":
                fake = generator(noise)
            elif generator.label_type == "required":
                fake = generator(noise, labels)
            else:
                fake = generator(noise, label_gen)
            eps = torch.rand(1).item()
            interpolate = eps * real_inputs + (1 - eps) * fake
            if discriminator.label_type == "none":
                d_interpolate = discriminator(interpolate)
            else:
                if generator.label_type == "generated":
                    d_interpolate = discriminator(interpolate, label_gen)
                else:
                    d_interpolate = discriminator(interpolate, labels)
            loss = self.forward(interpolate, d_interpolate)
            weighted_loss = self.lambd * loss
            weighted_loss.backward()
            optimizer_discriminator.step()
            return loss.item()



import torch

from .functional import least_squares_discriminator_loss, least_squares_generator_loss
from .loss import DiscriminatorLoss, GeneratorLoss

__all__ = ["LeastSquaresGeneratorLoss", "LeastSquaresDiscriminatorLoss"]


class LeastSquaresGeneratorLoss(GeneratorLoss):
    r"""Least Squares GAN generator loss from `"Least Squares Generative Adversarial Networks
    by Mao et. al." <https://arxiv.org/abs/1611.04076>`_ paper

    The loss can be described as

    .. math:: L(G) = \frac{(D(G(z)) - c)^2}{2}

    where

    - :math:`G` : Generator
    - :math:`D` : Disrciminator
    - :math:`c` : target generator label
    - :math:`z` : A sample from the noise prior

    Args:
        reduction (str, optional): Specifies the reduction to apply to the output.
            If ``none`` no reduction will be applied. If ``mean`` the outputs are averaged over batch size.
            If ``sum`` the elements of the output are summed.
        c (float, optional): Target generator label.
        override_train_ops (function, optional): Function to be used in place of the default ``train_ops``
    """

    def __init__(self, reduction="mean", c=1.0, override_train_ops=None):
        super(LeastSquaresGeneratorLoss, self).__init__(reduction, override_train_ops)
        self.c = c

    def forward(self, dgz):
        r"""Computes the loss for the given input.

        Args:
            dgz (torch.Tensor) : Output of the Discriminator with generated data. It must have the
                                 dimensions (N, \*) where \* means any number of additional
                                 dimensions.

        Returns:
            scalar if reduction is applied else Tensor with dimensions (N, \*).
        """
        return least_squares_generator_loss(dgz, self.c, self.reduction)


class LeastSquaresDiscriminatorLoss(DiscriminatorLoss):
    r"""Least Squares GAN discriminator loss from `"Least Squares Generative Adversarial Networks
    by Mao et. al." <https://arxiv.org/abs/1611.04076>`_ paper.

    The loss can be described as:

    .. math:: L(D) = \frac{(D(x) - b)^2 + (D(G(z)) - a)^2}{2}

    where

    - :math:`G` : Generator
    - :math:`D` : Disrciminator
    - :math:`a` : Target discriminator label for generated image
    - :math:`b` : Target discriminator label for real image

    Args:
        reduction (str, optional): Specifies the reduction to apply to the output.
            If ``none`` no reduction will be applied. If ``mean`` the outputs are averaged over batch size.
            If ``sum`` the elements of the output are summed.
        a (float, optional): Target discriminator label for generated image.
        b (float, optional): Target discriminator label for real image.
        override_train_ops (function, optional): Function to be used in place of the default ``train_ops``
    """

    def __init__(self, reduction="mean", a=0.0, b=1.0, override_train_ops=None):
        super(LeastSquaresDiscriminatorLoss, self).__init__(
            reduction, override_train_ops
        )
        self.a = a
        self.b = b

    def forward(self, dx, dgz):
        r"""Computes the loss for the given input.

        Args:
            dx (torch.Tensor) : Output of the Discriminator with real data. It must have the
                                dimensions (N, \*) where \* means any number of additional
                                dimensions.
            dgz (torch.Tensor) : Output of the Discriminator with generated data. It must have the
                                 dimensions (N, \*) where \* means any number of additional
                                 dimensions.

        Returns:
            scalar if reduction is applied else Tensor with dimensions (N, \*).
        """
        return least_squares_discriminator_loss(dx, dgz, self.a, self.b, self.reduction)



import torch

from ..models import AutoEncodingDiscriminator
from .functional import (
    energy_based_discriminator_loss,
    energy_based_generator_loss,
    energy_based_pulling_away_term,
)
from .loss import DiscriminatorLoss, GeneratorLoss

__all__ = [
    "EnergyBasedGeneratorLoss",
    "EnergyBasedDiscriminatorLoss",
    "EnergyBasedPullingAwayTerm",
]


class EnergyBasedGeneratorLoss(GeneratorLoss):
    r"""Energy Based GAN generator loss from `"Energy Based Generative Adversarial Network
    by Zhao et. al." <https://arxiv.org/abs/1609.03126>`_ paper.

    The loss can be described as:

    .. math:: L(G) = D(G(z))

    where

    - :math:`G` : Generator
    - :math:`D` : Discriminator
    - :math:`z` : A sample from the noise prior

    Args:
        reduction (str, optional): Specifies the reduction to apply to the output.
            If ``none`` no reduction will be applied. If ``mean`` the outputs are averaged over batch size.
            If ``sum`` the elements of the output are summed.
        override_train_ops (function, optional): A function is passed to this argument,
            if the default ``train_ops`` is not to be used.
    """

    def forward(self, dgz):
        r"""Computes the loss for the given input.

        Args:
            dgz (torch.Tensor): Output of the Discriminator with generated data. It must have the
                                dimensions (N, \*) where \* means any number of additional
                                dimensions.

        Returns:
            scalar if reduction is applied else Tensor with dimensions (N, \*).
        """
        return energy_based_generator_loss(dgz, self.reduction)

    def train_ops(
        self,
        generator,
        discriminator,
        optimizer_generator,
        device,
        batch_size,
        labels=None,
    ):
        r"""This function sets the ``embeddings`` attribute of the ``AutoEncodingDiscriminator`` to
        ``False`` and calls the ``train_ops`` of the ``GeneratorLoss``. After the call the
        attribute is again set to ``True``.

        Args:
            generator (torchgan.models.Generator): The model to be optimized.
            discriminator (torchgan.models.Discriminator): The discriminator which judges the
                performance of the generator.
            optimizer_generator (torch.optim.Optimizer): Optimizer which updates the ``parameters``
                of the ``generator``.
            device (torch.device): Device on which the ``generator`` and ``discriminator`` is present.
            batch_size (int): Batch Size of the data infered from the ``DataLoader`` by the ``Trainer``.
            labels (torch.Tensor, optional): Labels for the data.

        Returns:
            Scalar value of the loss.
        """
        if self.override_train_ops is not None:
            return self.override_train_ops(
                generator,
                discriminator,
                optimizer_generator,
                device,
                batch_size,
                labels,
            )
        else:
            if isinstance(discriminator, AutoEncodingDiscriminator):
                setattr(discriminator, "embeddings", False)
            loss = super(EnergyBasedGeneratorLoss, self).train_ops(
                generator,
                discriminator,
                optimizer_generator,
                device,
                batch_size,
                labels,
            )
            if isinstance(discriminator, AutoEncodingDiscriminator):
                setattr(discriminator, "embeddings", True)
            return loss


class EnergyBasedPullingAwayTerm(GeneratorLoss):
    r"""Energy Based Pulling Away Term from `"Energy Based Generative Adversarial Network
    by Zhao et. al." <https://arxiv.org/abs/1609.03126>`_ paper.

    The loss can be described as:

    .. math:: f_{PT}(S) = \frac{1}{N(N-1)}\sum_i\sum_{j \neq i}\bigg(\frac{S_i^T S_j}{||S_i||\ ||S_j||}\bigg)^2

    where

    - :math:`S` : The feature output from the encoder for generated images
    - :math:`N` : Batch Size of the Input

    Args:
        pt_ratio (float, optional): The weight given to the pulling away term.
        override_train_ops (function, optional): A function is passed to this argument,
            if the default ``train_ops`` is not to be used.
    """

    def __init__(self, pt_ratio=0.1, override_train_ops=None):
        super(EnergyBasedPullingAwayTerm, self).__init__("mean", override_train_ops)
        self.pt_ratio = pt_ratio

    def forward(self, dgz, d_hid):
        r"""Computes the loss for the given input.

        Args:
            dgz (torch.Tensor) : Output of the Discriminator with generated data. It must have the
                                 dimensions (N, \*) where \* means any number of additional
                                 dimensions.
            d_hid (torch.Tensor): The embeddings generated by the discriminator.

        Returns:
            scalar.
        """
        return self.pt_ratio * energy_based_pulling_away_term(d_hid)

    def train_ops(
        self,
        generator,
        discriminator,
        optimizer_generator,
        device,
        batch_size,
        labels=None,
    ):
        r"""This function extracts the hidden embeddings of the discriminator network. The furthur
        computation is same as the standard train_ops.

        .. note::
            For the loss to work properly, the discriminator must be a ``AutoEncodingDiscriminator``
            and it must have a ``embeddings`` attribute which should be set to ``True``. Also the
            generator ``label_type`` must be ``none``. As a result of these constraints it advisable
            not to use custom models with this loss. This will be improved in future.

        Args:
            generator (torchgan.models.Generator): The model to be optimized.
            discriminator (torchgan.models.Discriminator): The discriminator which judges the
                performance of the generator.
            optimizer_generator (torch.optim.Optimizer): Optimizer which updates the ``parameters``
                of the ``generator``.
            device (torch.device): Device on which the ``generator`` and ``discriminator`` is present.
            batch_size (int): Batch Size of the data infered from the ``DataLoader`` by the ``Trainer``.
            labels (torch.Tensor, optional): Labels for the data.

        Returns:
            Scalar value of the loss.
        """
        if self.override_train_ops is not None:
            return self.override_train_ops(
                generator,
                discriminator,
                optimizer_generator,
                device,
                batch_size,
                labels,
            )
        else:
            if not isinstance(discriminator, AutoEncodingDiscriminator):
                raise Exception(
                    "EBGAN PT requires the Discriminator to be a AutoEncoder"
                )
            if not generator.label_type == "none":
                raise Exception("EBGAN PT supports models which donot require labels")
            if not discriminator.embeddings:
                raise Exception("EBGAN PT requires the embeddings for loss computation")
            noise = torch.randn(batch_size, generator.encoding_dims, device=device)
            optimizer_generator.zero_grad()
            fake = generator(noise)
            d_hid, dgz = discriminator(fake)
            loss = self.forward(dgz, d_hid)
            loss.backward()
            optimizer_generator.step()
            return loss.item()


class EnergyBasedDiscriminatorLoss(DiscriminatorLoss):
    r"""Energy Based GAN generator loss from `"Energy Based Generative Adversarial Network
    by Zhao et. al." <https://arxiv.org/abs/1609.03126>`_ paper

    The loss can be described as:

    .. math:: L(D) = D(x) + max(0, m - D(G(z)))

    where

    - :math:`G` : Generator
    - :math:`D` : Discriminator
    - :math:`m` : Margin Hyperparameter
    - :math:`z` : A sample from the noise prior

    .. note::
        The convergence of EBGAN is highly sensitive to hyperparameters. The ``margin``
        hyperparameter as per the paper was taken as follows:

        +----------------------+--------+
        | Dataset              | Margin |
        +======================+========+
        | MNIST                | 10.0   |
        +----------------------+--------+
        | LSUN                 | 80.0   |
        +----------------------+--------+
        | CELEB A              | 20.0   |
        +----------------------+--------+
        | Imagenet (128 x 128) | 40.0   |
        +----------------------+--------+
        | Imagenet (256 x 256) | 80.0   |
        +----------------------+--------+

    Args:
        reduction (str, optional): Specifies the reduction to apply to the output.
            If ``none`` no reduction will be applied. If ``mean`` the outputs are averaged over batch size.
            If ``sum`` the elements of the output are summed.
        margin (float, optional): The margin hyperparameter.
        override_train_ops (function, optional): Function to be used in place of the default ``train_ops``
    """

    def __init__(self, reduction="mean", margin=80.0, override_train_ops=None):
        super(EnergyBasedDiscriminatorLoss, self).__init__(
            reduction, override_train_ops
        )
        self.margin = margin

    def forward(self, dx, dgz):
        r"""Computes the loss for the given input.

        Args:
            dx (torch.Tensor): Output of the Discriminator with real data. It must have the
                               dimensions (N, \*) where \* means any number of additional
                               dimensions.
            dgz (torch.Tensor): Output of the Discriminator with generated data. It must have the
                                dimensions (N, \*) where \* means any number of additional
                                dimensions.

        Returns:
            scalar if reduction is applied else Tensor with dimensions (N, \*).
        """
        return energy_based_discriminator_loss(dx, dgz, self.margin, self.reduction)

    def train_ops(
        self,
        generator,
        discriminator,
        optimizer_discriminator,
        real_inputs,
        device,
        batch_size,
        labels=None,
    ):
        r"""This function sets the ``embeddings`` attribute of the ``AutoEncodingDiscriminator`` to
        ``False`` and calls the ``train_ops`` of the ``DiscriminatorLoss``. After the call the
        attribute is again set to ``True``.

        Args:
            generator (torchgan.models.Generator): The model to be optimized.
            discriminator (torchgan.models.Discriminator): The discriminator which judges the
                performance of the generator.
            optimizer_discriminator (torch.optim.Optimizer): Optimizer which updates the ``parameters``
                of the ``discriminator``.
            real_inputs (torch.Tensor): The real data to be fed to the ``discriminator``.
            device (torch.device): Device on which the ``generator`` and ``discriminator`` is present.
            batch_size (int): Batch Size of the data infered from the ``DataLoader`` by the ``Trainer``.
            labels (torch.Tensor, optional): Labels for the data.

        Returns:
            Scalar value of the loss.
        """
        if self.override_train_ops is not None:
            return self.override_train_ops(
                self,
                generator,
                discriminator,
                optimizer_discriminator,
                real_inputs,
                device,
                labels,
            )
        else:
            if isinstance(discriminator, AutoEncodingDiscriminator):
                setattr(discriminator, "embeddings", False)
            loss = super(EnergyBasedDiscriminatorLoss, self).train_ops(
                generator,
                discriminator,
                optimizer_discriminator,
                real_inputs,
                device,
                labels,
            )
            if isinstance(discriminator, AutoEncodingDiscriminator):
                setattr(discriminator, "embeddings", True)
            return loss



import os
import sys
import unittest

import torch
from torchgan.layers import *

sys.path.append(os.path.join(os.path.dirname(__file__), "../.."))


class TestLayers(unittest.TestCase):
    def match_layer_outputs(self, layer, input, output_shape):
        z = layer(input)

        self.assertEqual(z.shape, output_shape)

    def test_residual_block2d(self):
        input = torch.rand(16, 3, 10, 10)

        layer = ResidualBlock2d([3, 16, 32, 3], [3, 3, 1], paddings=[1, 1, 0])

        self.match_layer_outputs(layer, input, (16, 3, 10, 10))

        layer = ResidualBlock2d(
            [3, 16, 32, 1],
            [3, 3, 1],
            paddings=[1, 1, 0],
            shortcut=torch.nn.Conv2d(3, 1, 3, padding=1),
        )

        self.match_layer_outputs(layer, input, (16, 1, 10, 10))

        layer = ResidualBlock2d([3, 16, 3], [1, 1])

        self.match_layer_outputs(layer, input, (16, 3, 10, 10))

    def test_transposed_residula_block2d(self):
        input = torch.rand(16, 3, 10, 10)

        layer = ResidualBlockTranspose2d([3, 16, 32, 3], [3, 3, 1], paddings=[1, 1, 0])

        self.match_layer_outputs(layer, input, (16, 3, 10, 10))

        layer = ResidualBlockTranspose2d(
            [3, 16, 32, 1],
            [3, 3, 1],
            paddings=[1, 1, 0],
            shortcut=torch.nn.Conv2d(3, 1, 3, padding=1),
        )

        self.match_layer_outputs(layer, input, (16, 1, 10, 10))

        layer = ResidualBlockTranspose2d([3, 16, 3], [1, 1])

        self.match_layer_outputs(layer, input, (16, 3, 10, 10))

    def test_basic_block2d(self):
        input = torch.rand(16, 3, 10, 10)

        layer = BasicBlock2d(3, 13, 3, 1, 1)

        self.match_layer_outputs(layer, input, (16, 16, 10, 10))

        layer = BasicBlock2d(3, 13, 3, 1, 1, batchnorm=False)

        self.match_layer_outputs(layer, input, (16, 16, 10, 10))

    def test_bottleneck_block2d(self):
        input = torch.rand(16, 3, 10, 10)

        layer = BottleneckBlock2d(3, 13, 3, 1, 1)

        self.match_layer_outputs(layer, input, (16, 16, 10, 10))

        layer = BottleneckBlock2d(3, 13, 3, 1, 1, batchnorm=False)

        self.match_layer_outputs(layer, input, (16, 16, 10, 10))

    def test_transition_block2d(self):
        input = torch.rand(16, 3, 10, 10)

        layer = TransitionBlock2d(3, 16, 3, 1, 1)

        self.match_layer_outputs(layer, input, (16, 16, 10, 10))

        layer = TransitionBlock2d(3, 16, 3, 1, 1, batchnorm=False)

        self.match_layer_outputs(layer, input, (16, 16, 10, 10))

    def test_transition_block_transpose2d(self):
        input = torch.rand(16, 3, 10, 10)

        layer = TransitionBlockTranspose2d(3, 16, 3, 1, 1)

        self.match_layer_outputs(layer, input, (16, 16, 10, 10))

        layer = TransitionBlockTranspose2d(3, 16, 3, 1, 1, batchnorm=False)

        self.match_layer_outputs(layer, input, (16, 16, 10, 10))

    def test_dense_block2d(self):
        input = torch.rand(16, 3, 10, 10)

        layer = DenseBlock2d(5, 3, 16, BottleneckBlock2d, 3, padding=1)

        self.match_layer_outputs(layer, input, (16, 83, 10, 10))

    def test_self_attention2d(self):
        input = torch.rand(16, 88, 10, 10)

        layer = SelfAttention2d(88)

        self.match_layer_outputs(layer, input, (16, 88, 10, 10))

    def test_spectral_norm2d(self):
        input = torch.rand(16, 3, 10, 10)

        layer = SpectralNorm2d(
            torch.nn.Conv2d(3, 10, 3, padding=1), power_iterations=10
        )

        self.match_layer_outputs(layer, input, (16, 10, 10, 10))



import torch

from .functional import mutual_information_penalty
from .loss import DiscriminatorLoss, GeneratorLoss

__all__ = ["MutualInformationPenalty"]


class MutualInformationPenalty(GeneratorLoss, DiscriminatorLoss):
    r"""Mutual Information Penalty as defined in
    `"InfoGAN : Interpretable Representation Learning by Information Maximising Generative Adversarial Nets
    by Chen et. al." <https://arxiv.org/abs/1606.03657>`_ paper

    The loss is the variational lower bound of the mutual information between
    the latent codes and the generator distribution and is defined as

    .. math:: L(G,Q) = log(Q|x)

    where

    - :math:`x` is drawn from the generator distribution G(z,c)
    - :math:`c` drawn from the latent code prior :math:`P(c)`

    Args:
        lambd (float, optional): The scaling factor for the loss.
        reduction (str, optional): Specifies the reduction to apply to the output.
            If ``none`` no reduction will be applied. If ``mean`` the mean of the output.
            If ``sum`` the elements of the output will be summed.
        override_train_ops (function, optional): A function is passed to this argument,
            if the default ``train_ops`` is not to be used.
    """

    def __init__(self, lambd=1.0, reduction="mean", override_train_ops=None):
        super(MutualInformationPenalty, self).__init__(reduction, override_train_ops)
        self.lambd = lambd

    def forward(self, c_dis, c_cont, dist_dis, dist_cont):
        r"""Computes the loss for the given input.

        Args:
            c_dis (int): The discrete latent code sampled from the prior.
            c_cont (int): The continuous latent code sampled from the prior.
            dist_dis (torch.distributions.Distribution): The auxilliary distribution :math:`Q(c|x)` over the
                discrete latent code output by the discriminator.
            dist_cont (torch.distributions.Distribution): The auxilliary distribution :math:`Q(c|x)` over the
                continuous latent code output by the discriminator.

        Returns:
            scalar if reduction is applied else Tensor with dimensions (N, \*).
        """
        return mutual_information_penalty(
            c_dis, c_cont, dist_dis, dist_cont, reduction=self.reduction
        )

    def train_ops(
        self,
        generator,
        discriminator,
        optimizer_generator,
        optimizer_discriminator,
        dis_code,
        cont_code,
        device,
        batch_size,
    ):
        if self.override_train_ops is not None:
            self.override_train_ops(
                generator,
                discriminator,
                optimizer_generator,
                optimizer_discriminator,
                dis_code,
                cont_code,
                device,
                batch_size,
            )
        else:
            noise = torch.randn(batch_size, generator.encoding_dims, device=device)
            optimizer_discriminator.zero_grad()
            optimizer_generator.zero_grad()
            fake = generator(noise, dis_code, cont_code)
            _, dist_dis, dist_cont = discriminator(fake, True)
            loss = self.forward(dis_code, cont_code, dist_dis, dist_cont)
            weighted_loss = self.lambd * loss
            weighted_loss.backward()
            optimizer_discriminator.step()
            optimizer_generator.step()
            return weighted_loss.item()



import torch
import torch.nn as nn

__all__ = ["VirtualBatchNorm"]


class VirtualBatchNorm(nn.Module):
    r"""Virtual Batch Normalization Module as proposed in the paper
    `"Improved Techniques for Training GANs by Salimans et. al." <https://arxiv.org/abs/1805.08318>`_

    Performs Normalizes the features of a batch based on the statistics collected on a reference
    batch of samples that are chosen once and fixed from the start, as opposed to regular
    batch normalization that uses the statistics of the batch being normalized

    Virtual Batch Normalization requires that the size of the batch being normalized is at least
    a multiple of (and ideally equal to) the size of the reference batch. Keep this in mind while
    choosing the batch size in ```torch.utils.data.DataLoader``` or use ```drop_last=True```

    .. math:: y = \frac{x - \mathrm{E}[x_{ref}]}{\sqrt{\mathrm{Var}[x_{ref}] + \epsilon}} * \gamma + \beta

    where

    - :math:`x` : Batch Being Normalized
    - :math:`x_{ref}` : Reference Batch

    Args:
        in_features (int): Size of the input dimension to be normalized
        eps (float, optional): Value to be added to variance for numerical stability while normalizing
    """

    def __init__(self, in_features, eps=1e-5):
        super(VirtualBatchNorm, self).__init__()
        self.in_features = in_features
        self.scale = nn.Parameter(torch.ones(in_features))
        self.bias = nn.Parameter(torch.zeros(in_features))
        self.ref_mu = None
        self.ref_var = None
        self.eps = eps

    def _batch_stats(self, x):
        r"""Computes the statistics of the batch ``x``.

        Args:
            x (torch.Tensor): Tensor whose statistics need to be computed.

        Returns:
            A tuple of the mean and variance of the batch ``x``.
        """
        mu = torch.mean(x, dim=0, keepdim=True)
        var = torch.var(x, dim=0, keepdim=True)
        return mu, var

    def _normalize(self, x, mu, var):
        r"""Normalizes the tensor ``x`` using the statistics ``mu`` and ``var``.

        Args:
            x (torch.Tensor): The Tensor to be normalized.
            mu (torch.Tensor): Mean using which the Tensor is to be normalized.
            var (torch.Tensor): Variance used in the normalization of ``x``.

        Returns:
            Normalized Tensor ``x``.
        """
        std = torch.sqrt(self.eps + var)
        x = (x - mu) / std
        sizes = list(x.size())
        for dim, i in enumerate(x.size()):
            if dim != 1:
                sizes[dim] = 1
        scale = self.scale.view(*sizes)
        bias = self.bias.view(*sizes)
        return x * scale + bias

    def forward(self, x):
        r"""Computes the output of the Virtual Batch Normalization

        Args:
            x (torch.Tensor): A Torch Tensor of dimension at least 2 which is to be Normalized

        Returns:
            Torch Tensor of the same dimension after normalizing with respect to the statistics of the reference batch
        """
        assert x.size(1) == self.in_features
        if self.ref_mu is None or self.ref_var is None:
            self.ref_mu, self.ref_var = self._batch_stats(x)
            self.ref_mu = self.ref_mu.clone().detach()
            self.ref_var = self.ref_var.clone().detach()
            out = self._normalize(x, self.ref_mu, self.ref_var)
        else:
            out = self._normalize(x, self.ref_mu, self.ref_var)
            self.ref_mu = None
            self.ref_var = None
        return out



import torch

from ..logging.logger import Logger
from ..losses.loss import DiscriminatorLoss, GeneratorLoss
from ..models.model import Discriminator, Generator
from .base_trainer import BaseTrainer

__all__ = ["Trainer"]


class Trainer(BaseTrainer):
    r"""Standard Trainer for various GANs. This has been designed to work only on one GPU in case
    you are using a GPU.

    Most of the functionalities provided by the Trainer are flexible enough and can be customized by
    simply passing different arguments. You can train anything from a simple DCGAN to complex CycleGANs
    without ever having to subclass this ``Trainer``.

    Args:
        models (dict): A dictionary containing a mapping between the variable name, storing the
            ``generator``, ``discriminator`` and any other model that you might want to define, with the
            function and arguments that are needed to construct the model. Refer to the examples to
            see how to define complex models using this API.
        losses_list (list): A list of the Loss Functions that need to be minimized. For a list of
            pre-defined losses look at :mod:`torchgan.losses`. All losses in the list must be a
            subclass of atleast ``GeneratorLoss`` or ``DiscriminatorLoss``.
        metrics_list (list, optional): List of Metric Functions that need to be logged. For a list of
            pre-defined metrics look at :mod:`torchgan.metrics`. All losses in the list must be a
            subclass of ``EvaluationMetric``.
        device (torch.device, optional): Device in which the operation is to be carried out. If you
            are using a CPU machine make sure that you change it for proper functioning.
        ncritic (int, optional): Setting it to a value will make the discriminator train that many
            times more than the generator. If it is set to a negative value the generator will be
            trained that many times more than the discriminator.
        sample_size (int, optional): Total number of images to be generated at the end of an epoch
            for logging purposes.
        epochs (int, optional): Total number of epochs for which the models are to be trained.
        checkpoints (str, optional): Path where the models are to be saved. The naming convention is
            if checkpoints is ``./model/gan`` then models are saved as ``./model/gan0.model`` and so on.
        retain_checkpoints (int, optional): Total number of checkpoints that should be retained. For
            example, if the value is set to 3, we save at most 3 models and start rewriting the models
            after that.
        recon (str, optional): Directory where the sampled images are saved. Make sure the directory
            exists from beforehand.
        log_dir (str, optional): The directory for logging tensorboard. It is ignored if
            TENSORBOARD_LOGGING is 0.
        test_noise (torch.Tensor, optional): If provided then it will be used as the noise for image
            sampling.
        nrow (int, optional): Number of rows in which the image is to be stored.

    Any other argument that you need to store in the object can be simply passed via keyword arguments.

    Example:
        >>> dcgan = Trainer(
                    {"generator": {"name": DCGANGenerator, "args": {"out_channels": 1, "step_channels":
                                   16}, "optimizer": {"name": Adam, "args": {"lr": 0.0002,
                                   "betas": (0.5, 0.999)}}},
                     "discriminator": {"name": DCGANDiscriminator, "args": {"in_channels": 1,
                                       "step_channels": 16}, "optimizer": {"var": "opt_discriminator",
                                       "name": Adam, "args": {"lr": 0.0002, "betas": (0.5, 0.999)}}}},
                    [MinimaxGeneratorLoss(), MinimaxDiscriminatorLoss()],
                    sample_size=64, epochs=20)
    """

    def __init__(
        self,
        models,
        losses_list,
        metrics_list=None,
        device=torch.device("cuda:0"),
        ncritic=1,
        epochs=5,
        sample_size=8,
        checkpoints="./model/gan",
        retain_checkpoints=5,
        recon="./images",
        log_dir=None,
        test_noise=None,
        nrow=8,
        **kwargs
    ):
        super(Trainer, self).__init__(
            losses_list,
            metrics_list=metrics_list,
            device=device,
            ncritic=ncritic,
            epochs=epochs,
            sample_size=sample_size,
            checkpoints=checkpoints,
            retain_checkpoints=retain_checkpoints,
            recon=recon,
            log_dir=log_dir,
            test_noise=test_noise,
            nrow=nrow,
            **kwargs
        )
        self.model_names = []
        self.optimizer_names = []
        self.schedulers = []
        for key, model in models.items():
            self.model_names.append(key)
            if "args" in model:
                setattr(self, key, (model["name"](**model["args"])).to(self.device))
            else:
                setattr(self, key, (model["name"]()).to(self.device))
            opt = model["optimizer"]
            opt_name = "optimizer_{}".format(key)
            if "var" in opt:
                opt_name = opt["var"]
            self.optimizer_names.append(opt_name)
            model_params = getattr(self, key).parameters()
            if "args" in opt:
                setattr(self, opt_name, (opt["name"](model_params, **opt["args"])))
            else:
                setattr(self, opt_name, (opt["name"](model_params)))
            if "scheduler" in opt:
                sched = opt["scheduler"]
                if "args" in sched:
                    self.schedulers.append(
                        sched["name"](getattr(self, opt_name), **sched["args"])
                    )
                else:
                    self.schedulers.append(sched["name"](getattr(self, opt_name)))

        self.logger = Logger(
            self,
            losses_list,
            metrics_list,
            log_dir=log_dir,
            nrow=nrow,
            test_noise=test_noise,
        )

        self._store_loss_maps()
        self._store_metric_maps()



import torch

from .functional import auxiliary_classification_loss
from .loss import DiscriminatorLoss, GeneratorLoss

__all__ = ["AuxiliaryClassifierGeneratorLoss", "AuxiliaryClassifierDiscriminatorLoss"]


class AuxiliaryClassifierGeneratorLoss(GeneratorLoss):
    r"""Auxiliary Classifier GAN (ACGAN) loss based on a from
    `"Conditional Image Synthesis With Auxiliary Classifier GANs
    by Odena et. al. " <https://arxiv.org/abs/1610.09585>`_ paper

    Args:
       reduction (str, optional): Specifies the reduction to apply to the output.
            If ``none`` no reduction will be applied. If ``mean`` the outputs are averaged over batch size.
            If ``sum`` the elements of the output are summed.
        override_train_ops (function, optional): A function is passed to this argument,
            if the default ``train_ops`` is not to be used.
    """

    def forward(self, logits, labels):
        return auxiliary_classification_loss(logits, labels, self.reduction)

    def train_ops(
        self,
        generator,
        discriminator,
        optimizer_generator,
        device,
        batch_size,
        labels=None,
    ):
        r"""Defines the standard ``train_ops`` used by the Auxiliary Classifier generator loss.

        The ``standard optimization algorithm`` for the ``discriminator`` defined in this train_ops
        is as follows (label_g and label_d both could be either real labels or generated labels):

        1. :math:`fake = generator(noise, label_g)`
        2. :math:`value_1 = classifier(fake, label_g)`
        3. :math:`value_2 = classifier(real, label_d)`
        4. :math:`loss = loss\_function(value_1, label_g) + loss\_function(value_2, label_d)`
        5. Backpropagate by computing :math:`\nabla loss`
        6. Run a step of the optimizer for discriminator

        Args:
            generator (torchgan.models.Generator): The model to be optimized. For ACGAN, it must require
                                                   labels for training
            discriminator (torchgan.models.Discriminator): The discriminator which judges the
                performance of the generator.
            optimizer_discriminator (torch.optim.Optimizer): Optimizer which updates the ``parameters``
                of the ``discriminator``.
            real_inputs (torch.Tensor): The real data to be fed to the ``discriminator``.
            device (torch.device): Device on which the ``generator`` and ``discriminator`` is present.
            batch_size (int): Batch Size of the data infered from the ``DataLoader`` by the ``Trainer``.
            labels (torch.Tensor, optional): Labels for the data.

        Returns:
            Scalar value of the loss.
        """
        if self.override_train_ops is not None:
            return self.override_train_ops(
                generator,
                discriminator,
                optimizer_generator,
                device,
                batch_size,
                labels,
            )
        if generator.label_type == "required" and labels is None:
            raise Exception("GAN model requires label for training")
        noise = torch.randn(batch_size, generator.encoding_dims, device=device)
        optimizer_generator.zero_grad()
        if generator.label_type == "none":
            raise Exception("Incorrect Model: ACGAN generator must require labels")
        if generator.label_type == "required":
            fake = generator(noise, labels)
        elif generator.label_type == "generated":
            label_gen = torch.randint(
                0, generator.num_classes, (batch_size,), device=device
            )
            fake = generator(noise, label_gen)
        cgz = discriminator(fake, mode="classifier")
        if generator.label_type == "required":
            loss = self.forward(cgz, labels)
        else:
            label_gen = label_gen.type(torch.LongTensor).to(device)
            loss = self.forward(cgz, label_gen)
        loss.backward()
        optimizer_generator.step()
        return loss.item()


class AuxiliaryClassifierDiscriminatorLoss(DiscriminatorLoss):
    r"""Auxiliary Classifier GAN (ACGAN) loss based on a from
    `"Conditional Image Synthesis With Auxiliary Classifier GANs
    by Odena et. al. " <https://arxiv.org/abs/1610.09585>`_ paper

    Args:
       reduction (str, optional): Specifies the reduction to apply to the output.
            If ``none`` no reduction will be applied. If ``mean`` the outputs are averaged over batch size.
            If ``sum`` the elements of the output are summed.
       override_train_ops (function, optional): A function is passed to this argument,
            if the default ``train_ops`` is not to be used.
    """

    def forward(self, logits, labels):
        return auxiliary_classification_loss(logits, labels, self.reduction)

    def train_ops(
        self,
        generator,
        discriminator,
        optimizer_discriminator,
        real_inputs,
        device,
        labels=None,
    ):
        r"""Defines the standard ``train_ops`` used by the Auxiliary Classifier discriminator loss.

        The ``standard optimization algorithm`` for the ``discriminator`` defined in this train_ops
        is as follows (label_g and label_d both could be either real labels or generated labels):

        1. :math:`fake = generator(noise, label_g)`
        2. :math:`value_1 = classifier(fake, label_g)`
        3. :math:`value_2 = classifier(real, label_d)`
        4. :math:`loss = loss\_function(value_1, label_g) + loss\_function(value_2, label_d)`
        5. Backpropagate by computing :math:`\nabla loss`
        6. Run a step of the optimizer for discriminator

        Args:
            generator (torchgan.models.Generator): The model to be optimized. For ACGAN, it must require labels
                                                   for training
            discriminator (torchgan.models.Discriminator): The discriminator which judges the
                performance of the generator.
            optimizer_discriminator (torch.optim.Optimizer): Optimizer which updates the ``parameters``
                of the ``discriminator``.
            real_inputs (torch.Tensor): The real data to be fed to the ``discriminator``.
            device (torch.device): Device on which the ``generator`` and ``discriminator`` is present.
            batch_size (int): Batch Size of the data infered from the ``DataLoader`` by the ``Trainer``.
            labels (torch.Tensor, optional): Labels for the data.

        Returns:
            Scalar value of the loss.
        """
        if self.override_train_ops is not None:
            return self.override_train_ops(
                generator,
                discriminator,
                optimizer_discriminator,
                real_inputs,
                device,
                labels,
            )
        if labels is None:
            raise Exception("ACGAN Discriminator requires labels for training")
        if generator.label_type == "none":
            raise Exception(
                "Incorrect Model: ACGAN generator must require labels for training"
            )
        batch_size = real_inputs.size(0)
        noise = torch.randn(batch_size, generator.encoding_dims, device=device)
        optimizer_discriminator.zero_grad()
        cx = discriminator(real_inputs, mode="classifier")
        if generator.label_type == "required":
            fake = generator(noise, labels)
        elif generator.label_type == "generated":
            label_gen = torch.randint(
                0, generator.num_classes, (batch_size,), device=device
            )
            fake = generator(noise, label_gen)
        cgz = discriminator(fake, mode="classifier")
        if generator.label_type == "required":
            loss = self.forward(cgz, labels) + self.forward(cx, labels)
        else:
            label_gen = label_gen.type(torch.LongTensor).to(device)
            loss = self.forward(cgz, label_gen) + self.forward(cx, labels)
        loss.backward()
        optimizer_discriminator.step()
        return loss.item()



from math import ceil, log2

import torch.nn as nn
import torch.nn.functional as F

from .model import Discriminator, Generator

__all__ = ["DCGANGenerator", "DCGANDiscriminator"]


class DCGANGenerator(Generator):
    r"""Deep Convolutional GAN (DCGAN) generator from
    `"Unsupervised Representation Learning With Deep Convolutional Generative Aversarial Networks
    by Radford et. al. " <https://arxiv.org/abs/1511.06434>`_ paper

    Args:
        encoding_dims (int, optional): Dimension of the encoding vector sampled from the noise prior.
        out_size (int, optional): Height and width of the input image to be generated. Must be at
            least 16 and should be an exact power of 2.
        out_channels (int, optional): Number of channels in the output Tensor.
        step_channels (int, optional): Number of channels in multiples of which the DCGAN steps up
            the convolutional features. The step up is done as dim :math:`z \rightarrow d \rightarrow
            2 \times d \rightarrow 4 \times d \rightarrow 8 \times d` where :math:`d` = step_channels.
        batchnorm (bool, optional): If True, use batch normalization in the convolutional layers of
            the generator.
        nonlinearity (torch.nn.Module, optional): Nonlinearity to be used in the intermediate
            convolutional layers. Defaults to ``LeakyReLU(0.2)`` when None is passed.
        last_nonlinearity (torch.nn.Module, optional): Nonlinearity to be used in the final
            convolutional layer. Defaults to ``Tanh()`` when None is passed.
        label_type (str, optional): The type of labels expected by the Generator. The available
            choices are 'none' if no label is needed, 'required' if the original labels are
            needed and 'generated' if labels are to be sampled from a distribution.
    """

    def __init__(
        self,
        encoding_dims=100,
        out_size=32,
        out_channels=3,
        step_channels=64,
        batchnorm=True,
        nonlinearity=None,
        last_nonlinearity=None,
        label_type="none",
    ):
        super(DCGANGenerator, self).__init__(encoding_dims, label_type)
        if out_size < 16 or ceil(log2(out_size)) != log2(out_size):
            raise Exception(
                "Target Image Size must be at least 16*16 and an exact power of 2"
            )
        num_repeats = out_size.bit_length() - 4
        self.ch = out_channels
        self.n = step_channels
        use_bias = not batchnorm
        nl = nn.LeakyReLU(0.2) if nonlinearity is None else nonlinearity
        last_nl = nn.Tanh() if last_nonlinearity is None else last_nonlinearity
        model = []
        d = int(self.n * (2 ** num_repeats))
        if batchnorm is True:
            model.append(
                nn.Sequential(
                    nn.ConvTranspose2d(self.encoding_dims, d, 4, 1, 0, bias=use_bias),
                    nn.BatchNorm2d(d),
                    nl,
                )
            )
            for i in range(num_repeats):
                model.append(
                    nn.Sequential(
                        nn.ConvTranspose2d(d, d // 2, 4, 2, 1, bias=use_bias),
                        nn.BatchNorm2d(d // 2),
                        nl,
                    )
                )
                d = d // 2
        else:
            model.append(
                nn.Sequential(
                    nn.ConvTranspose2d(self.encoding_dims, d, 4, 1, 0, bias=use_bias),
                    nl,
                )
            )
            for i in range(num_repeats):
                model.append(
                    nn.Sequential(
                        nn.ConvTranspose2d(d, d // 2, 4, 2, 1, bias=use_bias), nl
                    )
                )
                d = d // 2

        model.append(
            nn.Sequential(nn.ConvTranspose2d(d, self.ch, 4, 2, 1, bias=True), last_nl)
        )
        self.model = nn.Sequential(*model)
        self._weight_initializer()

    def forward(self, x, feature_matching=False):
        r"""Calculates the output tensor on passing the encoding ``x`` through the Generator.

        Args:
            x (torch.Tensor): A 2D torch tensor of the encoding sampled from a probability
                distribution.
            feature_matching (bool, optional): Returns the activation from a predefined intermediate
                layer.

        Returns:
            A 4D torch.Tensor of the generated image.
        """
        x = x.view(-1, x.size(1), 1, 1)
        return self.model(x)


class DCGANDiscriminator(Discriminator):
    r"""Deep Convolutional GAN (DCGAN) discriminator from
    `"Unsupervised Representation Learning With Deep Convolutional Generative Aversarial Networks
    by Radford et. al. " <https://arxiv.org/abs/1511.06434>`_ paper

    Args:
        in_size (int, optional): Height and width of the input image to be evaluated. Must be at
            least 16 and should be an exact power of 2.
        in_channels (int, optional): Number of channels in the input Tensor.
        step_channels (int, optional): Number of channels in multiples of which the DCGAN steps up
            the convolutional features. The step up is done as dim :math:`z \rightarrow d \rightarrow
            2 \times d \rightarrow 4 \times d \rightarrow 8 \times d` where :math:`d` = step_channels.
        batchnorm (bool, optional): If True, use batch normalization in the convolutional layers of
            the generator.
        nonlinearity (torch.nn.Module, optional): Nonlinearity to be used in the intermediate
            convolutional layers. Defaults to ``LeakyReLU(0.2)`` when None is passed.
        last_nonlinearity (torch.nn.Module, optional): Nonlinearity to be used in the final
            convolutional layer. Defaults to ``Tanh()`` when None is passed.
        label_type (str, optional): The type of labels expected by the Generator. The available
            choices are 'none' if no label is needed, 'required' if the original labels are
            needed and 'generated' if labels are to be sampled from a distribution.
    """

    def __init__(
        self,
        in_size=32,
        in_channels=3,
        step_channels=64,
        batchnorm=True,
        nonlinearity=None,
        last_nonlinearity=None,
        label_type="none",
    ):
        super(DCGANDiscriminator, self).__init__(in_channels, label_type)
        if in_size < 16 or ceil(log2(in_size)) != log2(in_size):
            raise Exception(
                "Input Image Size must be at least 16*16 and an exact power of 2"
            )
        num_repeats = in_size.bit_length() - 4
        self.n = step_channels
        use_bias = not batchnorm
        nl = nn.LeakyReLU(0.2) if nonlinearity is None else nonlinearity
        last_nl = nn.LeakyReLU(0.2) if last_nonlinearity is None else last_nonlinearity
        d = self.n
        model = [nn.Sequential(nn.Conv2d(self.input_dims, d, 4, 2, 1, bias=True), nl)]
        if batchnorm is True:
            for i in range(num_repeats):
                model.append(
                    nn.Sequential(
                        nn.Conv2d(d, d * 2, 4, 2, 1, bias=use_bias),
                        nn.BatchNorm2d(d * 2),
                        nl,
                    )
                )
                d *= 2
        else:
            for i in range(num_repeats):
                model.append(
                    nn.Sequential(nn.Conv2d(d, d * 2, 4, 2, 1, bias=use_bias), nl)
                )
                d *= 2
        self.disc = nn.Sequential(nn.Conv2d(d, 1, 4, 1, 0, bias=use_bias), last_nl)
        self.model = nn.Sequential(*model)
        self._weight_initializer()

    def forward(self, x, feature_matching=False):
        r"""Calculates the output tensor on passing the image ``x`` through the Discriminator.

        Args:
            x (torch.Tensor): A 4D torch tensor of the image.
            feature_matching (bool, optional): Returns the activation from a predefined intermediate
                layer.

        Returns:
            A 1D torch.Tensor of the probability of each image being real.
        """
        x = self.model(x)
        if feature_matching is True:
            return x
        else:
            x = self.disc(x)
            return x.view(x.size(0))



from pkgutil import iter_modules

import torch


def reduce(x, reduction=None):
    r"""Applies reduction on a torch.Tensor.

    Args:
        x (torch.Tensor): The tensor on which reduction is to be applied.
        reduction (str, optional): The reduction to be applied. If ``mean`` the  mean value of the
            Tensor is returned. If ``sum`` the elements of the Tensor will be summed. If none of the
            above then the Tensor is returning without any change.

    Returns:
        As per the above ``reduction`` convention.
    """
    if reduction == "mean":
        return torch.mean(x)
    elif reduction == "sum":
        return torch.sum(x)
    else:
        return x


def getenv_defaults(module_name):
    r"""Determines if a particular package is installed in the system.

    Args:
        module_name (str): The name of the package to be found.

    Returns:
        1 if package is installed else 0
    """
    return int(module_name in (name for loader, name, ispkg in iter_modules()))



import torch
import torch.autograd as autograd
import torch.nn.functional as F

from ..utils import reduce

__all__ = [
    "minimax_generator_loss",
    "minimax_discriminator_loss",
    "least_squares_generator_loss",
    "least_squares_discriminator_loss",
    "mutual_information_penalty",
    "wasserstein_generator_loss",
    "wasserstein_discriminator_loss",
    "wasserstein_gradient_penalty",
    "dragan_gradient_penalty",
    "auxiliary_classification_loss",
    "energy_based_generator_loss",
    "energy_based_discriminator_loss",
    "energy_based_pulling_away_term",
    "boundary_equilibrium_generator_loss",
    "boundary_equilibrium_discriminator_loss",
]

# Minimax Losses


def minimax_generator_loss(dgz, nonsaturating=True, reduction="mean"):
    if nonsaturating:
        target = torch.ones_like(dgz)
        return F.binary_cross_entropy_with_logits(dgz, target, reduction=reduction)
    else:
        target = torch.zeros_like(dgz)
        return -1.0 * F.binary_cross_entropy_with_logits(
            dgz, target, reduction=reduction
        )


def minimax_discriminator_loss(dx, dgz, label_smoothing=0.0, reduction="mean"):
    target_ones = torch.ones_like(dgz) * (1.0 - label_smoothing)
    target_zeros = torch.zeros_like(dx)
    loss = F.binary_cross_entropy_with_logits(dx, target_ones, reduction=reduction)
    loss += F.binary_cross_entropy_with_logits(dgz, target_zeros, reduction=reduction)
    return loss


# Least Squared Losses


def least_squares_generator_loss(dgz, c=1.0, reduction="mean"):
    return 0.5 * reduce((dgz - c) ** 2, reduction)


def least_squares_discriminator_loss(dx, dgz, a=0.0, b=1.0, reduction="mean"):
    return 0.5 * (reduce((dx - b) ** 2, reduction) + reduce((dgz - a) ** 2, reduction))


# Mutual Information Penalty


def mutual_information_penalty(c_dis, c_cont, dist_dis, dist_cont, reduction="mean"):
    log_probs = torch.Tensor(
        [
            torch.mean(dist.log_prob(c))
            for dist, c in zip((dist_dis, dist_cont), (c_dis, c_cont))
        ]
    )
    return reduce(-1.0 * log_probs, reduction)


# Wasserstein Losses


def wasserstein_generator_loss(fgz, reduction="mean"):
    return reduce(-1.0 * fgz, reduction)


def wasserstein_discriminator_loss(fx, fgz, reduction="mean"):
    return reduce(fgz - fx, reduction)


def wasserstein_gradient_penalty(interpolate, d_interpolate, reduction="mean"):
    grad_outputs = torch.ones_like(d_interpolate)
    gradients = autograd.grad(
        outputs=d_interpolate,
        inputs=interpolate,
        grad_outputs=grad_outputs,
        create_graph=True,
        retain_graph=True,
        only_inputs=True,
    )[0]

    gradient_penalty = (gradients.norm(2) - 1) ** 2
    return reduce(gradient_penalty, reduction)


# Dragan Penalty


def dragan_gradient_penalty(interpolate, d_interpolate, k=1.0, reduction="mean"):
    grad_outputs = torch.ones_like(d_interpolate)
    gradients = autograd.grad(
        outputs=d_interpolate,
        inputs=interpolate,
        grad_outputs=grad_outputs,
        create_graph=True,
        retain_graph=True,
        only_inputs=True,
        allow_unused=True,
    )[0]

    gradient_penalty = (gradients.norm(2) - k) ** 2
    return reduce(gradient_penalty, reduction)


# Auxiliary Classifier Loss


def auxiliary_classification_loss(logits, labels, reduction="mean"):
    return F.cross_entropy(logits, labels, reduction=reduction)


# Energy Based Losses


def energy_based_generator_loss(dgz, reduction="mean"):
    return reduce(dgz, reduction)


def energy_based_discriminator_loss(dx, dgz, margin, reduction="mean"):
    return reduce(dx + F.relu(-dgz + margin), reduction)


def energy_based_pulling_away_term(d_hid):
    d_hid_normalized = F.normalize(d_hid, p=2, dim=0)
    n = d_hid_normalized.size(0)
    d_hid_normalized = d_hid_normalized.view(n, -1)
    similarity = torch.matmul(d_hid_normalized, d_hid_normalized.transpose(1, 0))
    loss_pt = torch.sum(similarity ** 2) / (n * (n - 1))
    return loss_pt


# Boundary Equilibrium Losses


def boundary_equilibrium_generator_loss(dgz, reduction="mean"):
    return reduce(dgz, reduction)


def boundary_equilibrium_discriminator_loss(dx, dgz, k, reduction="mean"):
    # NOTE(avik-pal): This is a bit peculiar compared to the other losses as it must return 3 values.
    loss_real = reduce(dx, reduction)
    loss_fake = reduce(dgz, reduction)
    loss_total = loss_real - k * loss_fake
    return loss_total, loss_real, loss_fake



import os
import sys
import unittest

import torch
import torch.distributions as distributions
from torchgan.models import *

sys.path.append(os.path.join(os.path.dirname(__file__), "../.."))


class TestModels(unittest.TestCase):
    def test_dcgan_generator(self):
        encodings = [50, 100]
        channels = [3, 4]
        out_size = [32, 256]
        step = [64, 128]
        batchnorm = [True, False]
        nonlinearities = [None, torch.nn.ReLU()]
        last_nonlinearity = [None, torch.nn.LeakyReLU()]
        for i in range(2):
            gen = DCGANGenerator(
                encodings[i],
                out_size[i],
                channels[i],
                step[i],
                batchnorm[i],
                nonlinearities[i],
                last_nonlinearity[i],
            )
            z = torch.rand(10, encodings[i])
            x = gen(z)
            assert x.shape == (10, channels[i], out_size[i], out_size[i])

    def test_dcgan_discriminator(self):
        channels = [3, 4]
        step = [64, 128]
        in_size = [32, 64]
        batchnorm = [True, False]
        nonlinearities = [None, torch.nn.ReLU()]
        last_nonlinearity = [None, torch.nn.LeakyReLU()]
        for i in range(2):
            dis = DCGANDiscriminator(
                in_size[i],
                channels[i],
                step[i],
                batchnorm[i],
                nonlinearities[i],
                last_nonlinearity[i],
            )
            x = torch.rand(10, channels[i], in_size[i], in_size[i])
            loss = dis(x)
            assert loss.shape == (10,)

    def test_conditional_gan_generator(self):
        encodings = [50, 100]
        channels = [3, 4]
        out_size = [32, 64]
        classes = [5, 10]
        step = [64, 128]
        batchnorm = [True, False]
        nonlinearities = [None, torch.nn.ELU(0.5)]
        last_nonlinearity = [None, torch.nn.RReLU(0.25)]
        for i in range(2):
            ch = step[i]
            x = torch.randn(10, encodings[i])
            gen = ConditionalGANGenerator(
                classes[i],
                encodings[i],
                out_size[i],
                channels[i],
                ch,
                batchnorm[i],
                nonlinearities[i],
                last_nonlinearity[i],
            )
            labels = torch.randint(0, classes[i], (10,))
            y = gen(x, labels)
            assert y.shape == (10, channels[i], out_size[i], out_size[i])

    def test_conditional_gan_discriminator(self):
        channels = [3, 4]
        in_size = [32, 64]
        classes = [5, 10]
        step = [64, 128]
        batchnorm = [True, False]
        nonlinearities = [None, torch.nn.ELU(0.5)]
        last_nonlinearity = [None, torch.nn.RReLU(0.25)]
        for i in range(2):
            ch = step[i]
            x = torch.randn(10, channels[i], in_size[i], in_size[i])
            gen = ConditionalGANDiscriminator(
                classes[i],
                in_size[i],
                channels[i],
                ch,
                batchnorm[i],
                nonlinearities[i],
                last_nonlinearity[i],
            )
            labels = torch.randint(0, classes[i], (10,))
            y = gen(x, labels)
            assert y.shape == (10,)

    def test_acgan_generator(self):
        encodings = [50, 100]
        channels = [3, 4]
        out_size = [32, 64]
        classes = [5, 10]
        step = [64, 128]
        batchnorm = [True, False]
        nonlinearities = [None, torch.nn.ELU(0.5)]
        last_nonlinearity = [None, torch.nn.RReLU(0.25)]
        for i in range(2):
            ch = step[i]
            x = torch.randn(10, encodings[i])
            gen = ACGANGenerator(
                classes[i],
                encodings[i],
                out_size[i],
                channels[i],
                ch,
                batchnorm[i],
                nonlinearities[i],
                last_nonlinearity[i],
            )
            labels = torch.randint(0, classes[i], (10,))
            y = gen(x, labels)
            assert y.shape == (10, channels[i], out_size[i], out_size[i])

    def test_acgan_discriminator(self):
        channels = [3, 4]
        in_size = [32, 64]
        classes = [5, 10]
        step = [64, 128]
        batchnorm = [True, False]
        nonlinearities = [None, torch.nn.ELU(0.5)]
        last_nonlinearity = [None, torch.nn.RReLU(0.25)]
        for i in range(2):
            ch = step[i]
            x = torch.randn(10, channels[i], in_size[i], in_size[i])
            gen = ACGANDiscriminator(
                classes[i],
                in_size[i],
                channels[i],
                ch,
                batchnorm[i],
                nonlinearities[i],
                last_nonlinearity[i],
            )
            dx, cx = gen(x, mode="combine")
            assert dx.shape == (10,)
            assert cx.shape == (10, classes[i])

    def test_infogan_generator(self):
        encodings = [50, 100]
        channels = [3, 4]
        out_size = [32, 64]
        step = [64, 128]
        dim_cont = [10, 20]
        dim_dis = [30, 40]
        batchnorm = [True, False]
        nonlinearities = [None, torch.nn.ELU(0.5)]
        last_nonlinearity = [None, torch.nn.RReLU(0.25)]
        for i in range(2):
            ch = step[i]
            x = torch.randn(10, encodings[i])
            cont = torch.rand(10, dim_cont[i])
            dis = torch.zeros(10, dim_dis[i])
            gen = InfoGANGenerator(
                dim_cont[i],
                dim_dis[i],
                encodings[i],
                out_size[i],
                channels[i],
                ch,
                batchnorm[i],
                nonlinearities[i],
                last_nonlinearity[i],
            )
            y = gen(x, cont, dis)
            assert y.shape == (10, channels[i], out_size[i], out_size[i])

    def test_infogan_discriminator(self):
        channels = [3, 4]
        in_size = [32, 64]
        dim_cont = [10, 20]
        dim_dis = [30, 40]
        step = [64, 128]
        batchnorm = [True, False]
        nonlinearities = [None, torch.nn.ELU(0.5)]
        last_nonlinearity = [None, torch.nn.RReLU(0.25)]
        for i in range(2):
            x = torch.randn(10, channels[i], in_size[i], in_size[i])
            D = InfoGANDiscriminator(
                dim_dis[i],
                dim_cont[i],
                in_size[i],
                channels[i],
                step[i],
                batchnorm[i],
                nonlinearities[i],
                last_nonlinearity[i],
            )
            y, dist_dis, dist_cont = D(x, True)
            assert y.shape == (10, 1, 1, 1)
            assert isinstance(dist_dis, distributions.OneHotCategorical)
            assert isinstance(dist_cont, distributions.Normal)
            assert dist_dis.sample().shape == (10, dim_dis[i])
            assert dist_cont.sample().shape == (10, dim_cont[i])

    def test_autoencoding_generator(self):
        encodings = [50, 100]
        channels = [3, 4]
        out_size = [32, 64]
        step = [64, 128]
        scale = [2, 2]
        batchnorm = [True, False]
        nonlinearities = [None, torch.nn.ReLU()]
        last_nonlinearity = [None, torch.nn.LeakyReLU()]
        for i in range(2):
            gen = AutoEncodingGenerator(
                encodings[i],
                out_size[i],
                channels[i],
                step[i],
                scale[i],
                batchnorm[i],
                nonlinearities[i],
                last_nonlinearity[i],
            )
            z = torch.rand(10, encodings[i])
            x = gen(z)
            assert x.shape == (10, channels[i], out_size[i], out_size[i])

    def test_autoencoding_discriminator(self):
        channels = [3, 4]
        encodings = [50, 100]
        in_size = [32, 64]
        step = [64, 128]
        scale = [2, 2]
        batchnorm = [True, False]
        nonlinearities = [None, torch.nn.ReLU()]
        last_nonlinearity = [None, torch.nn.LeakyReLU()]
        for i in range(2):
            dis = AutoEncodingDiscriminator(
                in_size[i],
                channels[i],
                encodings[i],
                step[i],
                scale[i],
                batchnorm[i],
                nonlinearities[i],
                last_nonlinearity[i],
            )
            x = torch.rand(10, channels[i], in_size[i], in_size[i])
            loss = dis(x)
            assert loss.shape == (10,)



import torch
import torch.nn as nn
import torch.nn.functional as F

from .dcgan import DCGANDiscriminator, DCGANGenerator

__all__ = ["ACGANGenerator", "ACGANDiscriminator"]


class ACGANGenerator(DCGANGenerator):
    r"""Auxiliary Classifier GAN (ACGAN) generator based on a DCGAN model from
    `"Conditional Image Synthesis With Auxiliary Classifier GANs
    by Odena et. al. " <https://arxiv.org/abs/1610.09585>`_ paper

    Args:
        num_classes (int): Total classes present in the dataset.
        encoding_dims (int, optional): Dimension of the encoding vector sampled from the noise prior.
        out_size (int, optional): Height and width of the input image to be generated. Must be at
            least 16 and should be an exact power of 2.
        out_channels (int, optional): Number of channels in the output Tensor.
        step_channels (int, optional): Number of channels in multiples of which the DCGAN steps up
            the convolutional features. The step up is done as dim :math:`z \rightarrow d \rightarrow
            2 \times d \rightarrow 4 \times d \rightarrow 8 \times d` where :math:`d` = step_channels.
        batchnorm (bool, optional): If True, use batch normalization in the convolutional layers of
            the generator.
        nonlinearity (torch.nn.Module, optional): Nonlinearity to be used in the intermediate
            convolutional layers. Defaults to ``LeakyReLU(0.2)`` when None is passed.
        last_nonlinearity (torch.nn.Module, optional): Nonlinearity to be used in the final
            convolutional layer. Defaults to ``Tanh()`` when None is passed.
    """

    def __init__(
        self,
        num_classes,
        encoding_dims=100,
        out_size=32,
        out_channels=3,
        step_channels=64,
        batchnorm=True,
        nonlinearity=None,
        last_nonlinearity=None,
    ):
        super(ACGANGenerator, self).__init__(
            encoding_dims,
            out_size,
            out_channels,
            step_channels,
            batchnorm,
            nonlinearity,
            last_nonlinearity,
            label_type="generated",
        )
        self.encoding_dims = encoding_dims
        self.num_classes = num_classes
        self.label_embeddings = nn.Embedding(self.num_classes, self.encoding_dims)

    def forward(self, z, y):
        r"""Calculates the output tensor on passing the encoding ``z`` through the Generator.

        Args:
            z (torch.Tensor): A 2D torch tensor of the encoding sampled from a probability
                distribution.
            y (torch.Tensor): The labels corresponding to the encoding ``z``.

        Returns:
            A 4D torch.Tensor of the generated Images conditioned on ``y``.
        """
        y_emb = self.label_embeddings(y.type(torch.LongTensor).to(y.device))
        return super(ACGANGenerator, self).forward(torch.mul(y_emb, z))

    def sampler(self, sample_size, device):
        return [
            torch.randn(sample_size, self.encoding_dims, device=device),
            torch.randint(0, self.num_classes, (sample_size,), device=device),
        ]


class ACGANDiscriminator(DCGANDiscriminator):
    r"""Auxiliary Classifier GAN (ACGAN) discriminator based on a DCGAN model from
    `"Conditional Image Synthesis With Auxiliary Classifier GANs
    by Odena et. al. " <https://arxiv.org/abs/1610.09585>`_ paper

    Args:
        num_classes (int): Total classes present in the dataset.
        in_size (int, optional): Height and width of the input image to be evaluated. Must be at
            least 16 and should be an exact power of 2.
        in_channels (int, optional): Number of channels in the input Tensor.
        step_channels (int, optional): Number of channels in multiples of which the DCGAN steps up
            the convolutional features. The step up is done as dim :math:`z \rightarrow d \rightarrow
            2 \times d \rightarrow 4 \times d \rightarrow 8 \times d` where :math:`d` = step_channels.
        batchnorm (bool, optional): If True, use batch normalization in the convolutional layers of
            the generator.
        nonlinearity (torch.nn.Module, optional): Nonlinearity to be used in the intermediate
            convolutional layers. Defaults to ``LeakyReLU(0.2)`` when None is passed.
        last_nonlinearity (torch.nn.Module, optional): Nonlinearity to be used in the final
            convolutional layer. Defaults to ``Tanh()`` when None is passed.
    """

    def __init__(
        self,
        num_classes,
        in_size=32,
        in_channels=3,
        step_channels=64,
        batchnorm=True,
        nonlinearity=None,
        last_nonlinearity=None,
    ):
        super(ACGANDiscriminator, self).__init__(
            in_size,
            in_channels,
            step_channels,
            batchnorm,
            nonlinearity,
            last_nonlinearity,
            label_type="none",
        )
        last_nl = nn.LeakyReLU(0.2) if last_nonlinearity is None else last_nonlinearity
        self.input_dims = in_channels
        self.num_classes = num_classes
        d = self.n * 2 ** (in_size.bit_length() - 4)
        self.aux = nn.Sequential(
            nn.Conv2d(d, self.num_classes, 4, 1, 0, bias=False), last_nl
        )

    def forward(self, x, mode="discriminator", feature_matching=False):
        r"""Calculates the output tensor on passing the image ``x`` through the Discriminator.

        Args:
            x (torch.Tensor): A 4D torch tensor of the image.
            mode (str, optional): Option to choose the mode of the ACGANDiscriminator. Setting it to
                'discriminator' gives the probability of the image being fake/real, 'classifier' allows
                it to make a prediction about the class of the image and anything else leads to
                returning both the values.
            feature_matching (bool, optional): Returns the activation from a predefined intermediate
                layer.

        Returns:
            A 1D torch.Tensor of the probability of each image being real.
        """
        x = self.model(x)
        if feature_matching is True:
            return x
        if mode == "discriminator":
            dx = self.disc(x)
            return dx.view(dx.size(0))
        elif mode == "classifier":
            cx = self.aux(x)
            return cx.view(cx.size(0), cx.size(1))
        else:
            dx = self.disc(x)
            cx = self.aux(x)
            return dx.view(dx.size(0)), cx.view(cx.size(0), cx.size(1))



import torch
import torch.nn as nn

__all__ = [
    "BasicBlock2d",
    "BottleneckBlock2d",
    "TransitionBlock2d",
    "TransitionBlockTranspose2d",
    "DenseBlock2d",
]


class BasicBlock2d(nn.Module):
    r"""Basic Block Module as described in `"Densely Connected Convolutional Networks by Huang et.
    al." <https://arxiv.org/abs/1608.06993>`_

    The output is computed by ``concatenating`` the ``input`` tensor to the ``output`` tensor (of the
    internal model) along the ``channel`` dimension.

    The internal model is simply a sequence of a ``Conv2d`` layer and a ``BatchNorm2d`` layer, if
    activated.

    Args:
        in_channels (int): The channel dimension of the input tensor.
        out_channels (int): The channel dimension of the output tensor.
        kernel (int, tuple): Size of the Convolutional Kernel.
        stride (int, tuple, optional): Stride of the Convolutional Kernel.
        padding (int, tuple, optional): Padding to be applied on the input tensor.
        batchnorm (bool, optional): If ``True``, batch normalization shall be performed.
        nonlinearity (torch.nn.Module, optional): Activation to be applied. Defaults to
                                                  ``torch.nn.LeakyReLU``.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel,
        stride=1,
        padding=0,
        batchnorm=True,
        nonlinearity=None,
    ):
        super(BasicBlock2d, self).__init__()
        nl = nn.LeakyReLU(0.2) if nonlinearity is None else nonlinearity
        if batchnorm is True:
            self.model = nn.Sequential(
                nn.BatchNorm2d(in_channels),
                nl,
                nn.Conv2d(
                    in_channels, out_channels, kernel, stride, padding, bias=False
                ),
            )
        else:
            self.model = nn.Sequential(
                nl,
                nn.Conv2d(
                    in_channels, out_channels, kernel, stride, padding, bias=True
                ),
            )

    def forward(self, x):
        r"""Computes the output of the basic dense block

        Args:
            x (torch.Tensor): The input tensor having channel dimension same as ``in_channels``.

        Returns:
            4D Tensor by concatenating the input to the output of the internal model.
        """
        return torch.cat([x, self.model(x)], 1)


class BottleneckBlock2d(nn.Module):
    r"""Bottleneck Block Module as described in `"Densely Connected Convolutional Networks by Huang
    et. al." <https://arxiv.org/abs/1608.06993>`_

    The output is computed by ``concatenating`` the ``input`` tensor to the ``output`` tensor (of the
    internal model) along the ``channel`` dimension.

    The internal model is simply a sequence of 2 ``Conv2d`` layers and 2 ``BatchNorm2d`` layers, if
    activated. This Module is much more computationally efficient than the ``BasicBlock2d``, and hence
    is more recommended.

    Args:
        in_channels (int): The channel dimension of the input tensor.
        out_channels (int): The channel dimension of the output tensor.
        kernel (int, tuple): Size of the Convolutional Kernel.
        stride (int, tuple, optional): Stride of the Convolutional Kernel.
        padding (int, tuple, optional): Padding to be applied on the input tensor.
        bottleneck_channels (int, optional): The channels in the intermediate convolutional
                                             layer. A higher value will make learning of
                                             more complex functions possible. Defaults to
                                             ``4 * in_channels``.
        batchnorm (bool, optional): If ``True``, batch normalization shall be performed.
        nonlinearity (torch.nn.Module, optional): Activation to be applied. Defaults to
                                                  ``torch.nn.LeakyReLU``.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel,
        stride=1,
        padding=0,
        bottleneck_channels=None,
        batchnorm=True,
        nonlinearity=None,
    ):
        super(BottleneckBlock2d, self).__init__()
        bottleneck_channels = (
            4 * in_channels if bottleneck_channels is None else bottleneck_channels
        )
        nl = nn.LeakyReLU(0.2) if nonlinearity is None else nonlinearity
        if batchnorm is True:
            self.model = nn.Sequential(
                nn.BatchNorm2d(in_channels),
                nl,
                nn.Conv2d(in_channels, bottleneck_channels, 1, 1, 0, bias=False),
                nn.BatchNorm2d(bottleneck_channels),
                nl,
                nn.Conv2d(
                    bottleneck_channels,
                    out_channels,
                    kernel,
                    stride,
                    padding,
                    bias=False,
                ),
            )
        else:
            self.model = nn.Sequential(
                nl,
                nn.Conv2d(in_channels, bottleneck_channels, 1, 1, 0, bias=True),
                nl,
                nn.Conv2d(
                    bottleneck_channels,
                    out_channels,
                    kernel,
                    stride,
                    padding,
                    bias=True,
                ),
            )

    def forward(self, x):
        r"""Computes the output of the bottleneck dense block

        Args:
            x (torch.Tensor): The input tensor having channel dimension same as ``in_channels``.

        Returns:
            4D Tensor by concatenating the input to the output of the internal model.
        """
        return torch.cat([x, self.model(x)], 1)


class TransitionBlock2d(nn.Module):
    r"""Transition Block Module as described in `"Densely Connected Convolutional Networks by Huang
    et. al." <https://arxiv.org/abs/1608.06993>`_

    This is a simple ``Sequential`` model of a ``Conv2d`` layer and a ``BatchNorm2d`` layer, if
    activated.

    Args:
        in_channels (int): The channel dimension of the input tensor.
        out_channels (int): The channel dimension of the output tensor.
        kernel (int, tuple): Size of the Convolutional Kernel.
        stride (int, tuple, optional): Stride of the Convolutional Kernel.
        padding (int, tuple, optional): Padding to be applied on the input tensor.
        batchnorm (bool, optional): If ``True``, batch normalization shall be performed.
        nonlinearity (torch.nn.Module, optional): Activation to be applied. Defaults to
                                                  ``torch.nn.LeakyReLU``.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel,
        stride=1,
        padding=0,
        batchnorm=True,
        nonlinearity=None,
    ):
        super(TransitionBlock2d, self).__init__()
        nl = nn.LeakyReLU(0.2) if nonlinearity is None else nonlinearity
        if batchnorm is True:
            self.model = nn.Sequential(
                nn.BatchNorm2d(in_channels),
                nl,
                nn.Conv2d(
                    in_channels, out_channels, kernel, stride, padding, bias=False
                ),
            )
        else:
            self.model = nn.Sequential(
                nl,
                nn.Conv2d(
                    in_channels, out_channels, kernel, stride, padding, bias=True
                ),
            )

    def forward(self, x):
        r"""Computes the output of the transition block

        Args:
            x (torch.Tensor): The input tensor having channel dimension same as ``in_channels``.

        Returns:
            4D Tensor by applying the ``model`` on ``x``.
        """
        return self.model(x)


class TransitionBlockTranspose2d(nn.Module):
    r"""Transition Block Transpose Module is constructed by simply reversing the effect of
    Transition Block Module. We replace the ``Conv2d`` layers by ``ConvTranspose2d`` layers.

    Args:
        in_channels (int): The channel dimension of the input tensor.
        out_channels (int): The channel dimension of the output tensor.
        kernel (int, tuple): Size of the Convolutional Kernel.
        stride (int, tuple, optional): Stride of the Convolutional Kernel.
        padding (int, tuple, optional): Padding to be applied on the input tensor.
        batchnorm (bool, optional): If ``True``, batch normalization shall be performed.
        nonlinearity (torch.nn.Module, optional): Activation to be applied. Defaults to
                                                  ``torch.nn.LeakyReLU``.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel,
        stride=1,
        padding=0,
        batchnorm=True,
        nonlinearity=None,
    ):
        super(TransitionBlockTranspose2d, self).__init__()
        nl = nn.LeakyReLU(0.2) if nonlinearity is None else nonlinearity
        if batchnorm is True:
            self.model = nn.Sequential(
                nn.BatchNorm2d(in_channels),
                nl,
                nn.ConvTranspose2d(
                    in_channels, out_channels, kernel, stride, padding, bias=False
                ),
            )
        else:
            self.model = nn.Sequential(
                nl,
                nn.ConvTranspose2d(
                    in_channels, out_channels, kernel, stride, padding, bias=True
                ),
            )

    def forward(self, x):
        r"""Computes the output of the transition block transpose

        Args:
            x (torch.Tensor): The input tensor having channel dimension same as ``in_channels``.

        Returns:
            4D Tensor by applying the ``model`` on ``x``.
        """
        return self.model(x)


class DenseBlock2d(nn.Module):
    r"""Dense Block Module as described in `"Densely Connected Convolutional Networks by Huang
    et. al." <https://arxiv.org/abs/1608.06993>`_

    Args:
        depth (int): The total number of ``blocks`` that will be present.
        in_channels (int): The channel dimension of the input tensor.
        growth_rate (int): The rate at which the channel dimension increases. The output of
                           the module has a channel dimension of size ``in_channels +
                           depth * growth_rate``.
        block (torch.nn.Module): Should be once of the Densenet Blocks. Forms the building block
                                 for the Dense Block.
        kernel (int, tuple): Size of the Convolutional Kernel.
        stride (int, tuple, optional): Stride of the Convolutional Kernel.
        padding (int, tuple, optional): Padding to be applied on the input tensor.
        batchnorm (bool, optional): If ``True``, batch normalization shall be performed.
        nonlinearity (torch.nn.Module, optional): Activation to be applied. Defaults to
                                                  ``torch.nn.LeakyReLU``.
    """

    def __init__(
        self,
        depth,
        in_channels,
        growth_rate,
        block,
        kernel,
        stride=1,
        padding=0,
        batchnorm=True,
        nonlinearity=None,
    ):
        super(DenseBlock2d, self).__init__()
        nl = nn.LeakyReLU(0.2) if nonlinearity is None else nonlinearity
        model = []
        for i in range(depth):
            # FIXME(Aniket1998): There is no way to pass an option for bottleneck channels
            model.append(
                block(
                    in_channels + i * growth_rate,
                    growth_rate,
                    kernel,
                    stride,
                    padding,
                    batchnorm=batchnorm,
                    nonlinearity=nl,
                )
            )
        self.model = nn.Sequential(*model)

    def forward(self, x):
        r"""Computes the output of the transition block transpose

        Args:
            x (torch.Tensor): The input tensor having channel dimension same as ``in_channels``.

        Returns:
            4D Tensor by applying the ``model`` on ``x``.
        """
        return self.model(x)



import torch

from ..utils import reduce
from .loss import DiscriminatorLoss, GeneratorLoss

__all__ = ["HistoricalAverageGeneratorLoss", "HistoricalAverageDiscriminatorLoss"]


class HistoricalAverageGeneratorLoss(GeneratorLoss):
    r"""Historical Average Generator Loss from
    `"Improved Techniques for Training GANs
    by Salimans et. al." <https://arxiv.org/pdf/1606.03498.pdf>`_ paper

    The loss can be described as

    .. math:: || \vtheta - \frac{1}{t} \sum_{i=1}^t \vtheta[i] ||^2

    where

    - :math:`G` : Generator
    - :math: `\vtheta[i]` : Generator Parameters at Past Timestep :math: `i`

    Args:
        reduction (str, optional): Specifies the reduction to apply to the output.
            If ``none`` no reduction will be applied. If ``mean`` the outputs are averaged over batch size.
            If ``sum`` the elements of the output are summed.
        override_train_ops (function, optional): Function to be used in place of the default ``train_ops``
        lambd (float, optional): Hyperparameter lambda for scaling the Historical Average Penalty
    """

    def __init__(
        self, reduction="elementwise_mean", override_train_ops=None, lambd=1.0
    ):
        super(HistoricalAverageGeneratorLoss, self).__init__(
            reduction, override_train_ops
        )
        self.timesteps = 0
        self.sum_parameters = []
        self.lambd = lambd

    r"""Defines the standard ``train_ops`` used by historical averaging loss.

        The ``standard optimization algorithm`` for the ``discriminator`` defined in this train_ops
        is as follows:

        1. Compute the loss :math: `|| \vtheta - \frac{1}{t} \sum_{i=1}^t \vtheta[i] ||^2`
        2. Backpropagate by computing :math:`\nabla loss`
        3. Run a step of the optimizer for discriminator

        Args:
            generator (torchgan.models.Generator): The model to be optimized.
            optimizer_generator (torch.optim.Optimizer): Optimizer which updates the ``parameters``
                of the ``generator``.

        Returns:
            Scalar value of the loss.
    """

    def train_ops(self, generator, optimizer_generator):
        if self.override_train_ops is not None:
            return self.override_train_ops(self, generator, optimizer_generator)
        else:
            if self.timesteps == 0:
                for p in generator.parameters():
                    param = p.data.clone()
                    self.sum_parameters.append(param)
                self.timesteps += 1
                return 0.0
            else:
                optimizer_generator.zero_grad()
                loss = 0.0
                for i, p in enumerate(generator.parameters()):
                    loss += torch.sum(
                        (p - (self.sum_parameters[i].data / self.timesteps)) ** 2
                    )
                    self.sum_parameters[i] += p.data.clone()
                self.timesteps += 1
                loss *= self.lambd
                loss.backward()
                optimizer_generator.step()
                return loss.item()


class HistoricalAverageDiscriminatorLoss(DiscriminatorLoss):
    r"""Historical Average Discriminator Loss from
    `"Improved Techniques for Training GANs
    by Salimans et. al." <https://arxiv.org/pdf/1606.03498.pdf>`_ paper

    The loss can be described as

    .. math:: || \vtheta - \frac{1}{t} \sum_{i=1}^t \vtheta[i] ||^2

    where

    - :math:`G` : Discriminator
    - :math: `\vtheta[i]` : Discriminator Parameters at Past Timestep :math: `i`

    Args:
        reduction (str, optional): Specifies the reduction to apply to the output.
            If ``none`` no reduction will be applied. If ``mean`` the outputs are averaged over batch size.
            If ``sum`` the elements of the output are summed.
        override_train_ops (function, optional): Function to be used in place of the default ``train_ops``
        lambd (float, optional): Hyperparameter lambda for scaling the Historical Average Penalty
    """

    def __init__(
        self, reduction="elementwise_mean", override_train_ops=None, lambd=1.0
    ):
        super(HistoricalAverageDiscriminatorLoss, self).__init__(
            reduction, override_train_ops
        )
        self.timesteps = 0
        self.sum_parameters = []
        self.lambd = lambd

    r"""Defines the standard ``train_ops`` used by historical averaging loss.

        The ``standard optimization algorithm`` for the ``discriminator`` defined in this train_ops
        is as follows:

        1. Compute the loss :math: `|| \vtheta - \frac{1}{t} \sum_{i=1}^t \vtheta[i] ||^2`
        2. Backpropagate by computing :math:`\nabla loss`
        3. Run a step of the optimizer for discriminator

        Args:
            generator (torchgan.models.Generator): The model to be optimized.
            optimizer_generator (torch.optim.Optimizer): Optimizer which updates the ``parameters``
                of the ``generator``.

        Returns:
            Scalar value of the loss.
    """

    def train_ops(self, discriminator, optimizer_discriminator):
        if self.override_train_ops is not None:
            return self.override_train_ops(self, discriminator, optimizer_discriminator)
        else:
            if self.timesteps == 0:
                for p in discriminator.parameters():
                    param = p.data.clone()
                    self.sum_parameters.append(param)
                self.timesteps += 1
                return 0.0
            else:
                optimizer_discriminator.zero_grad()
                loss = 0.0
                for i, p in enumerate(discriminator.parameters()):
                    loss += torch.sum(
                        (p - (self.sum_parameters[i].data / self.timesteps)) ** 2
                    )
                    self.sum_parameters[i] += p.data.clone()
                self.timesteps += 1
                loss *= self.lambd
                loss.backward()
                optimizer_discriminator.step()
                return loss.item()



import torch.nn as nn

__all__ = ["ResidualBlock2d", "ResidualBlockTranspose2d"]


class ResidualBlock2d(nn.Module):
    r"""Residual Block Module as described in `"Deep Residual Learning for Image Recognition
    by He et. al." <https://arxiv.org/abs/1512.03385>`_

    The output of the residual block is computed in the following manner:

    .. math:: output = activation(layers(x) + shortcut(x))

    where

    - :math:`x` : Input to the Module
    - :math:`layers` : The feed forward network
    - :math:`shortcut` : The function to be applied along the skip connection
    - :math:`activation` : The activation function applied at the end of the residual block

    Args:
        filters (list): A list of the filter sizes. For ex, if the input has a channel
            dimension of 16, and you want 3 convolution layers and the final output to have a
            channel dimension of 16, then the list would be [16, 32, 64, 16].
        kernels (list): A list of the kernel sizes. Each kernel size can be an integer or a
            tuple, similar to Pytorch convention. The length of the ``kernels`` list must be
            1 less than the ``filters`` list.
        strides (list, optional): A list of the strides for each convolution layer.
        paddings (list, optional): A list of the padding in each convolution layer.
        nonlinearity (torch.nn.Module, optional): The activation to be used after every convolution
            layer.
        batchnorm (bool, optional): If set to ``False``, batch normalization is not used after
            every convolution layer.
        shortcut (torch.nn.Module, optional): The function to be applied on the input along the
            skip connection.
        last_nonlinearity (torch.nn.Module, optional): The activation to be applied at the end of
            the residual block.
    """

    def __init__(
        self,
        filters,
        kernels,
        strides=None,
        paddings=None,
        nonlinearity=None,
        batchnorm=True,
        shortcut=None,
        last_nonlinearity=None,
    ):
        super(ResidualBlock2d, self).__init__()
        nl = nn.LeakyReLU(0.2) if nonlinearity is None else nonlinearity
        if strides is None:
            strides = [1 for _ in range(len(kernels))]
        if paddings is None:
            paddings = [0 for _ in range(len(kernels))]
        assert (
            len(filters) == len(kernels) + 1
            and len(filters) == len(strides) + 1
            and len(filters) == len(paddings) + 1
        )
        layers = []
        for i in range(1, len(filters)):
            layers.append(
                nn.Conv2d(
                    filters[i - 1],
                    filters[i],
                    kernels[i - 1],
                    strides[i - 1],
                    paddings[i - 1],
                )
            )
            if batchnorm:
                layers.append(nn.BatchNorm2d(filters[i]))
            if i != len(filters) - 1:  # Last layer does not get an activation
                layers.append(nl)
        self.layers = nn.Sequential(*layers)
        self.shortcut = shortcut
        self.last_nonlinearity = last_nonlinearity

    def forward(self, x):
        r"""Computes the output of the residual block

        Args:
            x (torch.Tensor): A 4D Torch Tensor which is the input to the Residual Block.

        Returns:
            4D Torch Tensor after applying the desired functions as specified while creating the
            object.
        """
        out = self.layers(x)
        if self.shortcut is not None:
            out += self.shortcut(x)
        else:
            out += x
        return out if self.last_nonlinearity is None else self.last_nonlinearity(out)


class ResidualBlockTranspose2d(nn.Module):
    r"""A customized version of Residual Block having Conv Transpose layers instead of Conv layers.

    The output of this block is computed in the following manner:

    .. math:: output = activation(layers(x) + shortcut(x))

    where

    - :math:`x` : Input to the Module
    - :math:`layers` : The feed forward network
    - :math:`shortcut` : The function to be applied along the skip connection
    - :math:`activation` : The activation function applied at the end of the residual block

    Args:
        filters (list): A list of the filter sizes. For ex, if the input has a channel
            dimension of 16, and you want 3 transposed convolution layers and the final output
            to have a channel dimension of 16, then the list would be [16, 32, 64, 16].
        kernels (list): A list of the kernel sizes. Each kernel size can be an integer or a
            tuple, similar to Pytorch convention. The length of the ``kernels`` list must be
            1 less than the ``filters`` list.
        strides (list, optional): A list of the strides for each convolution layer.
        paddings (list, optional): A list of the padding in each convolution layer.
        nonlinearity (torch.nn.Module, optional): The activation to be used after every convolution
            layer.
        batchnorm (bool, optional): If set to ``False``, batch normalization is not used after
            every convolution layer.
        shortcut (torch.nn.Module, optional): The function to be applied on the input along the
            skip connection.
        last_nonlinearity (torch.nn.Module, optional): The activation to be applied at the end of
            the residual block.
    """

    def __init__(
        self,
        filters,
        kernels,
        strides=None,
        paddings=None,
        nonlinearity=None,
        batchnorm=True,
        shortcut=None,
        last_nonlinearity=None,
    ):
        super(ResidualBlockTranspose2d, self).__init__()
        nl = nn.LeakyReLU(0.2) if nonlinearity is None else nonlinearity
        if strides is None:
            strides = [1 for _ in range(len(kernels))]
        if paddings is None:
            paddings = [0 for _ in range(len(kernels))]
        assert (
            len(filters) == len(kernels) + 1
            and len(filters) == len(strides) + 1
            and len(filters) == len(paddings) + 1
        )
        layers = []
        for i in range(1, len(filters)):
            layers.append(
                nn.ConvTranspose2d(
                    filters[i - 1],
                    filters[i],
                    kernels[i - 1],
                    strides[i - 1],
                    paddings[i - 1],
                )
            )
            if batchnorm:
                layers.append(nn.BatchNorm2d(filters[i]))
            if i != len(filters) - 1:  # Last layer does not get an activation
                layers.append(nl)
        self.layers = nn.Sequential(*layers)
        self.shortcut = shortcut
        self.last_nonlinearity = last_nonlinearity

    def forward(self, x):
        r"""Computes the output of the residual block

        Args:
            x (torch.Tensor): A 4D Torch Tensor which is the input to the Transposed Residual Block.

        Returns:
            4D Torch Tensor after applying the desired functions as specified while creating the
            object.
        """
        out = self.layers(x)
        if self.shortcut is not None:
            out += self.shortcut(x)
        else:
            out += x
        return out if self.last_nonlinearity is None else self.last_nonlinearity(out)
